# Copyright 2013 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

# This module's classes provide an interface to mojo modules. Modules are
# collections of interfaces and structs to be used by mojo ipc clients and
# servers.
#
# A simple interface would be created this way:
# module = mojom.generate.module.Module('Foo')
# interface = module.AddInterface('Bar')
# method = interface.AddMethod('Tat', 0)
# method.AddParameter('baz', 0, mojom.INT32)

import pickle
from collections import OrderedDict
from uuid import UUID

# pylint: disable=raise-missing-from


class BackwardCompatibilityChecker:
  """Used for memoization while recursively checking two type definitions for
  backward-compatibility."""

  def __init__(self):
    self._cache = {}

  def IsBackwardCompatible(self, new_kind, old_kind):
    key = (new_kind, old_kind)
    result = self._cache.get(key)
    if result is None:
      # Assume they're compatible at first to effectively ignore recursive
      # checks between these types, e.g. if both kinds are a struct or union
      # that references itself in a field.
      self._cache[key] = True
      result = new_kind.IsBackwardCompatible(old_kind, self)
      self._cache[key] = result
    return result


# We use our own version of __repr__ when displaying the AST, as the
# AST currently doesn't capture which nodes are reference (e.g. to
# types) and which nodes are definitions. This allows us to e.g. print
# the definition of a struct when it's defined inside a module, but
# only print its name when it's referenced in e.g. a method parameter.
def Repr(obj, as_ref=True):
  """A version of __repr__ that can distinguish references.

  Sometimes we like to print an object's full representation
  (e.g. with its fields) and sometimes we just want to reference an
  object that was printed in full elsewhere. This function allows us
  to make that distinction.

  Args:
    obj: The object whose string representation we compute.
    as_ref: If True, use the short reference representation.

  Returns:
    A str representation of |obj|.
  """
  if hasattr(obj, 'Repr'):
    return obj.Repr(as_ref=as_ref)
  # Since we cannot implement Repr for existing container types, we
  # handle them here.
  if isinstance(obj, list):
    if not obj:
      return '[]'
    return ('[\n%s\n]' %
            (',\n'.join('    %s' % Repr(elem, as_ref).replace('\n', '\n    ')
                        for elem in obj)))
  if isinstance(obj, dict):
    if not obj:
      return '{}'
    return ('{\n%s\n}' % (',\n'.join('    %s: %s' %
                                     (Repr(key, as_ref).replace('\n', '\n    '),
                                      Repr(val, as_ref).replace('\n', '\n    '))
                                     for key, val in obj.items())))
  return repr(obj)


def GenericRepr(obj, names):
  """Compute generic Repr for |obj| based on the attributes in |names|.

  Args:
    obj: The object to compute a Repr for.
    names: A dict from attribute names to include, to booleans
        specifying whether those attributes should be shown as
        references or not.

  Returns:
    A str representation of |obj|.
  """

  def ReprIndent(name, as_ref):
    return '    %s=%s' % (name, Repr(getattr(obj, name), as_ref).replace(
        '\n', '\n    '))

  return '%s(\n%s\n)' % (obj.__class__.__name__, ',\n'.join(
      ReprIndent(name, as_ref) for (name, as_ref) in names.items()))


class Kind:
  """Kind represents a type (e.g. int8, string).

  Attributes:
    spec: A string uniquely identifying the type. May be None.
    module: {Module} The defining module. Set to None for built-in types.
    parent_kind: The enclosing type. For example, an enum defined
        inside an interface has that interface as its parent. May be None.
    is_nullable: True if the type is nullable.
  """

  def __init__(self, spec=None, is_nullable=False, module=None):
    self.spec = spec
    self.module = module
    self.parent_kind = None
    self.is_nullable = is_nullable
    self.shared_definition = {}

  @classmethod
  def AddSharedProperty(cls, name):
    """Adds a property |name| to |cls|, which accesses the corresponding item in
       |shared_definition|.

       The reason of adding such indirection is to enable sharing definition
       between a reference kind and its nullable variation. For example:
         a = Struct('test_struct_1')
         b = a.MakeNullableKind()
         a.name = 'test_struct_2'
         print(b.name)  # Outputs 'test_struct_2'.
    """
    def Get(self):
      try:
        return self.shared_definition[name]
      except KeyError:  # Must raise AttributeError if property doesn't exist.
        raise AttributeError

    def Set(self, value):
      self.shared_definition[name] = value

    setattr(cls, name, property(Get, Set))

  def Repr(self, as_ref=True):
    # pylint: disable=unused-argument
    return '<%s spec=%r is_nullable=%r>' % (self.__class__.__name__, self.spec,
                                            self.is_nullable)

  def __repr__(self):
    # Gives us a decent __repr__ for all kinds.
    return self.Repr()

  def __eq__(self, rhs):
    # pylint: disable=unidiomatic-typecheck
    return (type(self) == type(rhs)
            and (self.spec, self.parent_kind, self.is_nullable)
            == (rhs.spec, rhs.parent_kind, rhs.is_nullable))

  def __hash__(self):
    # TODO(crbug.com/1060471): Remove this and other __hash__ methods on Kind
    # and its subclasses. This is to support existing generator code which uses
    # some primitive Kinds as dict keys. The default hash (object identity)
    # breaks these dicts when a pickled Module instance is unpickled and used
    # during a subsequent run of the parser.
    return hash((self.spec, self.parent_kind, self.is_nullable))

  # pylint: disable=unused-argument
  def IsBackwardCompatible(self, rhs, checker):
    return self == rhs


class ValueKind(Kind):
  """ValueKind represents values that aren't reference kinds.

  The primary difference is the wire representation for nullable value kinds
  still reserves space for the value type itself, even if that value itself
  is logically null.
  """
  def __init__(self, spec=None, is_nullable=False, module=None):
    assert spec is None or is_nullable == spec.startswith('?')
    Kind.__init__(self, spec, is_nullable, module)

  def MakeNullableKind(self):
    assert not self.is_nullable

    if self == BOOL:
      return NULLABLE_BOOL
    if self == INT8:
      return NULLABLE_INT8
    if self == INT16:
      return NULLABLE_INT16
    if self == INT32:
      return NULLABLE_INT32
    if self == INT64:
      return NULLABLE_INT64
    if self == UINT8:
      return NULLABLE_UINT8
    if self == UINT16:
      return NULLABLE_UINT16
    if self == UINT32:
      return NULLABLE_UINT32
    if self == UINT64:
      return NULLABLE_UINT64
    if self == FLOAT:
      return NULLABLE_FLOAT
    if self == DOUBLE:
      return NULLABLE_DOUBLE

    nullable_kind = type(self)()
    nullable_kind.shared_definition = self.shared_definition
    if self.spec is not None:
      nullable_kind.spec = '?' + self.spec
    nullable_kind.is_nullable = True
    nullable_kind.parent_kind = self.parent_kind
    nullable_kind.module = self.module

    return nullable_kind

  def MakeUnnullableKind(self):
    assert self.is_nullable

    if self == NULLABLE_BOOL:
      return BOOL
    if self == NULLABLE_INT8:
      return INT8
    if self == NULLABLE_INT16:
      return INT16
    if self == NULLABLE_INT32:
      return INT32
    if self == NULLABLE_INT64:
      return INT64
    if self == NULLABLE_UINT8:
      return UINT8
    if self == NULLABLE_UINT16:
      return UINT16
    if self == NULLABLE_UINT32:
      return UINT32
    if self == NULLABLE_UINT64:
      return UINT64
    if self == NULLABLE_FLOAT:
      return FLOAT
    if self == NULLABLE_DOUBLE:
      return DOUBLE

    nullable_kind = type(self)()
    nullable_kind.shared_definition = self.shared_definition
    if self.spec is not None:
      nullable_kind.spec = self.spec[1:]
    nullable_kind.is_nullable = False
    nullable_kind.parent_kind = self.parent_kind
    nullable_kind.module = self.module

    return nullable_kind

  def __eq__(self, rhs):
    return (isinstance(rhs, ValueKind) and super().__eq__(rhs))

  def __hash__(self):  # pylint: disable=useless-super-delegation
    return super().__hash__()


class ReferenceKind(Kind):
  """ReferenceKind represents pointer and handle types.

  A type is nullable if null (for pointer types) or invalid handle (for handle
  types) is a legal value for the type.
  """

  def __init__(self, spec=None, is_nullable=False, module=None):
    assert spec is None or is_nullable == spec.startswith('?')
    Kind.__init__(self, spec, is_nullable, module)

  def MakeNullableKind(self):
    assert not self.is_nullable

    if self == STRING:
      return NULLABLE_STRING
    if self == HANDLE:
      return NULLABLE_HANDLE
    if self == DCPIPE:
      return NULLABLE_DCPIPE
    if self == DPPIPE:
      return NULLABLE_DPPIPE
    if self == MSGPIPE:
      return NULLABLE_MSGPIPE
    if self == SHAREDBUFFER:
      return NULLABLE_SHAREDBUFFER
    if self == PLATFORMHANDLE:
      return NULLABLE_PLATFORMHANDLE

    nullable_kind = type(self)()
    nullable_kind.shared_definition = self.shared_definition
    if self.spec is not None:
      nullable_kind.spec = '?' + self.spec
    nullable_kind.is_nullable = True
    nullable_kind.parent_kind = self.parent_kind
    nullable_kind.module = self.module

    return nullable_kind

  def MakeUnnullableKind(self):
    assert self.is_nullable

    if self == NULLABLE_STRING:
      return STRING
    if self == NULLABLE_HANDLE:
      return HANDLE
    if self == NULLABLE_DCPIPE:
      return DCPIPE
    if self == NULLABLE_DPPIPE:
      return DPPIPE
    if self == NULLABLE_MSGPIPE:
      return MSGPIPE
    if self == NULLABLE_SHAREDBUFFER:
      return SHAREDBUFFER
    if self == NULLABLE_PLATFORMHANDLE:
      return PLATFORMHANDLE

    unnullable_kind = type(self)()
    unnullable_kind.shared_definition = self.shared_definition
    if self.spec is not None:
      assert self.spec[0] == '?'
      unnullable_kind.spec = self.spec[1:]
    unnullable_kind.is_nullable = False
    unnullable_kind.parent_kind = self.parent_kind
    unnullable_kind.module = self.module

    return unnullable_kind

  def __eq__(self, rhs):
    return (isinstance(rhs, ReferenceKind) and super().__eq__(rhs))

  def __hash__(self):  # pylint: disable=useless-super-delegation
    return super().__hash__()


# Initialize the set of primitive types. These can be accessed by clients.
BOOL = ValueKind('b')
INT8 = ValueKind('i8')
INT16 = ValueKind('i16')
INT32 = ValueKind('i32')
INT64 = ValueKind('i64')
UINT8 = ValueKind('u8')
UINT16 = ValueKind('u16')
UINT32 = ValueKind('u32')
UINT64 = ValueKind('u64')
FLOAT = ValueKind('f')
DOUBLE = ValueKind('d')
NULLABLE_BOOL = ValueKind('?b', True)
NULLABLE_INT8 = ValueKind('?i8', True)
NULLABLE_INT16 = ValueKind('?i16', True)
NULLABLE_INT32 = ValueKind('?i32', True)
NULLABLE_INT64 = ValueKind('?i64', True)
NULLABLE_UINT8 = ValueKind('?u8', True)
NULLABLE_UINT16 = ValueKind('?u16', True)
NULLABLE_UINT32 = ValueKind('?u32', True)
NULLABLE_UINT64 = ValueKind('?u64', True)
NULLABLE_FLOAT = ValueKind('?f', True)
NULLABLE_DOUBLE = ValueKind('?d', True)
STRING = ReferenceKind('s')
HANDLE = ReferenceKind('h')
DCPIPE = ReferenceKind('h:d:c')
DPPIPE = ReferenceKind('h:d:p')
MSGPIPE = ReferenceKind('h:m')
SHAREDBUFFER = ReferenceKind('h:s')
PLATFORMHANDLE = ReferenceKind('h:p')
NULLABLE_STRING = ReferenceKind('?s', True)
NULLABLE_HANDLE = ReferenceKind('?h', True)
NULLABLE_DCPIPE = ReferenceKind('?h:d:c', True)
NULLABLE_DPPIPE = ReferenceKind('?h:d:p', True)
NULLABLE_MSGPIPE = ReferenceKind('?h:m', True)
NULLABLE_SHAREDBUFFER = ReferenceKind('?h:s', True)
NULLABLE_PLATFORMHANDLE = ReferenceKind('?h:p', True)

# Collection of all Primitive types
PRIMITIVES = (
    BOOL,
    INT8,
    INT16,
    INT32,
    INT64,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
    FLOAT,
    DOUBLE,
    NULLABLE_BOOL,
    NULLABLE_INT8,
    NULLABLE_INT16,
    NULLABLE_INT32,
    NULLABLE_INT64,
    NULLABLE_UINT8,
    NULLABLE_UINT16,
    NULLABLE_UINT32,
    NULLABLE_UINT64,
    NULLABLE_FLOAT,
    NULLABLE_DOUBLE,
    STRING,
    HANDLE,
    DCPIPE,
    DPPIPE,
    MSGPIPE,
    SHAREDBUFFER,
    PLATFORMHANDLE,
    NULLABLE_STRING,
    NULLABLE_HANDLE,
    NULLABLE_DCPIPE,
    NULLABLE_DPPIPE,
    NULLABLE_MSGPIPE,
    NULLABLE_SHAREDBUFFER,
    NULLABLE_PLATFORMHANDLE,
)

ATTRIBUTE_MIN_VERSION = 'MinVersion'
ATTRIBUTE_DEFAULT = 'Default'
ATTRIBUTE_EXTENSIBLE = 'Extensible'
ATTRIBUTE_NO_INTERRUPT = 'NoInterrupt'
ATTRIBUTE_STABLE = 'Stable'
ATTRIBUTE_SUPPORTS_URGENT = 'SupportsUrgent'
ATTRIBUTE_SYNC = 'Sync'
ATTRIBUTE_UNLIMITED_SIZE = 'UnlimitedSize'
ATTRIBUTE_UUID = 'Uuid'
ATTRIBUTE_SERVICE_SANDBOX = 'ServiceSandbox'
ATTRIBUTE_REQUIRE_CONTEXT = 'RequireContext'
ATTRIBUTE_ALLOWED_CONTEXT = 'AllowedContext'
ATTRIBUTE_RUNTIME_FEATURE = 'RuntimeFeature'


class NamedValue:
  def __init__(self, module, parent_kind, mojom_name):
    self.module = module
    self.parent_kind = parent_kind
    self.mojom_name = mojom_name

  def GetSpec(self):
    return (self.module.GetNamespacePrefix() +
            (self.parent_kind and
             (self.parent_kind.mojom_name + '.') or "") + self.mojom_name)

  def __eq__(self, rhs):
    return (isinstance(rhs, NamedValue)
            and (self.parent_kind, self.mojom_name) == (rhs.parent_kind,
                                                        rhs.mojom_name))

  def __hash__(self):
    return hash((self.parent_kind, self.mojom_name))


class BuiltinValue:
  def __init__(self, value):
    self.value = value

  def __eq__(self, rhs):
    return isinstance(rhs, BuiltinValue) and self.value == rhs.value


class ConstantValue(NamedValue):
  def __init__(self, module, parent_kind, constant):
    NamedValue.__init__(self, module, parent_kind, constant.mojom_name)
    self.constant = constant

  @property
  def name(self):
    return self.constant.name


class EnumValue(NamedValue):
  def __init__(self, module, enum, field):
    NamedValue.__init__(self, module, enum.parent_kind, field.mojom_name)
    self.field = field
    self.enum = enum

  def GetSpec(self):
    return (self.module.GetNamespacePrefix() +
            (self.parent_kind and (self.parent_kind.mojom_name + '.') or "") +
            self.enum.mojom_name + '.' + self.mojom_name)

  @property
  def name(self):
    return self.field.name


class Constant:
  def __init__(self, mojom_name=None, kind=None, value=None, parent_kind=None):
    self.mojom_name = mojom_name
    self.name = None
    self.kind = kind
    self.value = value
    self.parent_kind = parent_kind

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeConstant(self.mojom_name)

  def __eq__(self, rhs):
    return (isinstance(rhs, Constant)
            and (self.mojom_name, self.kind, self.value,
                 self.parent_kind) == (rhs.mojom_name, rhs.kind, rhs.value,
                                       rhs.parent_kind))


class Field:
  def __init__(self,
               mojom_name=None,
               kind=None,
               ordinal=None,
               default=None,
               attributes=None):
    if self.__class__.__name__ == 'Field':
      raise Exception()
    self.mojom_name = mojom_name
    self.name = None
    self.kind = kind
    self.ordinal = ordinal
    self.default = default
    self.attributes = attributes

  def Repr(self, as_ref=True):
    # pylint: disable=unused-argument
    # Fields are only referenced by objects which define them and thus
    # they are always displayed as non-references.
    return GenericRepr(self, {'mojom_name': False, 'kind': True})

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeField(self.mojom_name)

  @property
  def min_version(self):
    return self.attributes.get(ATTRIBUTE_MIN_VERSION) \
        if self.attributes else None

  def __eq__(self, rhs):
    return (isinstance(rhs, Field)
            and (self.mojom_name, self.kind, self.ordinal, self.default,
                 self.attributes) == (rhs.mojom_name, rhs.kind, rhs.ordinal,
                                      rhs.default, rhs.attributes))

  def __hash__(self):
    return hash((self.mojom_name, self.kind, self.ordinal, self.default))


class StructField(Field):
  def __hash__(self):
    return super(Field, self).__hash__()


class UnionField(Field):
  def __init__(self,
               mojom_name=None,
               kind=None,
               ordinal=None,
               default=None,
               attributes=None):
    Field.__init__(self, mojom_name, kind, ordinal, default, attributes)

  @property
  def is_default(self):
    return self.attributes.get(ATTRIBUTE_DEFAULT, False) \
        if self.attributes else False


def _IsFieldBackwardCompatible(new_field, old_field, checker):
  if (new_field.min_version or 0) != (old_field.min_version or 0):
    return False

  return checker.IsBackwardCompatible(new_field.kind, old_field.kind)


class Feature(ReferenceKind):
  """A runtime enabled feature defined from mojom.

  Attributes:
    mojom_name: {str} The name of the feature type as defined in mojom.
    name: {str} The stylized name. (Note: not the "name" used by FeatureList.)
    constants: {List[Constant]} The constants defined in the feature scope.
    attributes: {dict} Additional information about the feature.
  """

  Kind.AddSharedProperty('mojom_name')
  Kind.AddSharedProperty('name')
  Kind.AddSharedProperty('constants')
  Kind.AddSharedProperty('attributes')

  def __init__(self, mojom_name=None, module=None, attributes=None):
    if mojom_name is not None:
      spec = 'x:' + mojom_name
    else:
      spec = None
    ReferenceKind.__init__(self, spec, False, module)
    self.mojom_name = mojom_name
    self.name = None
    self.constants = []
    self.attributes = attributes

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeFeature(self.mojom_name)
    for constant in self.constants:
      constant.Stylize(stylizer)


class Struct(ReferenceKind):
  """A struct with typed fields.

  Attributes:
    mojom_name: {str} The name of the struct type as defined in mojom.
    name: {str} The stylized name.
    native_only: {bool} Does the struct have a body (i.e. any fields) or is it
        purely a native struct.
    custom_serializer: {bool} Should we generate a serializer for the struct or
        will one be provided by non-generated code.
    fields: {List[StructField]} The members of the struct.
    enums: {List[Enum]} The enums defined in the struct scope.
    constants: {List[Constant]} The constants defined in the struct scope.
    attributes: {dict} Additional information about the struct, such as
        if it's a native struct.
  """

  Kind.AddSharedProperty('mojom_name')
  Kind.AddSharedProperty('name')
  Kind.AddSharedProperty('native_only')
  Kind.AddSharedProperty('custom_serializer')
  Kind.AddSharedProperty('fields')
  Kind.AddSharedProperty('enums')
  Kind.AddSharedProperty('constants')
  Kind.AddSharedProperty('attributes')

  def __init__(self, mojom_name=None, module=None, attributes=None):
    if mojom_name is not None:
      spec = 'x:' + mojom_name
    else:
      spec = None
    ReferenceKind.__init__(self, spec, False, module)
    self.mojom_name = mojom_name
    self.name = None
    self.native_only = False
    self.custom_serializer = False
    self.fields = []
    self.enums = []
    self.constants = []
    self.attributes = attributes

  def Repr(self, as_ref=True):
    if as_ref:
      return '<%s mojom_name=%r module=%s>' % (self.__class__.__name__,
                                               self.mojom_name,
                                               Repr(self.module, as_ref=True))
    return GenericRepr(self, {
        'mojom_name': False,
        'fields': False,
        'module': True
    })

  def AddField(self,
               mojom_name,
               kind,
               ordinal=None,
               default=None,
               attributes=None):
    field = StructField(mojom_name, kind, ordinal, default, attributes)
    self.fields.append(field)
    return field

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeStruct(self.mojom_name)
    for field in self.fields:
      field.Stylize(stylizer)
    for enum in self.enums:
      enum.Stylize(stylizer)
    for constant in self.constants:
      constant.Stylize(stylizer)

  def IsBackwardCompatible(self, rhs, checker):
    """This struct is backward-compatible with rhs (older_struct) if and only if
    all of the following conditions hold:
      - Any newly added field is tagged with a [MinVersion] attribute specifying
        a version number greater than all previously used [MinVersion]
        attributes within the struct.
      - All fields present in rhs remain present in the new struct,
        with the same ordinal position, same optional or non-optional status,
        same (or backward-compatible) type and where applicable, the same
        [MinVersion] attribute value.
      - All [MinVersion] attributes must be non-decreasing in ordinal order.
      - All reference-typed (string, array, map, struct, or union) fields tagged
        with a [MinVersion] greater than zero must be optional.
    """

    def buildOrdinalFieldMap(struct):
      fields_by_ordinal = {}
      for field in struct.fields:
        if field.ordinal in fields_by_ordinal:
          raise Exception('Multiple fields with ordinal %s in struct %s.' %
                          (field.ordinal, struct.mojom_name))
        fields_by_ordinal[field.ordinal] = field
      return fields_by_ordinal

    new_fields = buildOrdinalFieldMap(self)
    old_fields = buildOrdinalFieldMap(rhs)
    if len(new_fields) < len(old_fields):
      # At least one field was removed, which is not OK.
      return False

    # If there are N fields, existing ordinal values must exactly cover the
    # range from 0 to N-1.
    num_old_ordinals = len(old_fields)
    max_old_min_version = 0
    for ordinal in range(num_old_ordinals):
      new_field = new_fields[ordinal]
      old_field = old_fields[ordinal]
      if (old_field.min_version or 0) > max_old_min_version:
        max_old_min_version = old_field.min_version
      if not _IsFieldBackwardCompatible(new_field, old_field, checker):
        # Type or min-version mismatch between old and new versions of the same
        # ordinal field.
        return False

    # At this point we know all old fields are intact in the new struct
    # definition. Now verify that all new fields have a high enough min version
    # and are appropriately optional where required.
    num_new_ordinals = len(new_fields)
    last_min_version = max_old_min_version
    for ordinal in range(num_old_ordinals, num_new_ordinals):
      new_field = new_fields[ordinal]
      min_version = new_field.min_version or 0
      if min_version <= max_old_min_version:
        # A new field is being added to an existing version, which is not OK.
        return False
      if min_version < last_min_version:
        # The [MinVersion] of a field cannot be lower than the [MinVersion] of
        # a field with lower ordinal value.
        return False
      if IsReferenceKind(new_field.kind) and not IsNullableKind(new_field.kind):
        # New fields whose type can be nullable MUST be nullable.
        return False

    return True

  @property
  def stable(self):
    return self.attributes.get(ATTRIBUTE_STABLE, False) \
        if self.attributes else False

  @property
  def qualified_name(self):
    if self.parent_kind:
      prefix = self.parent_kind.qualified_name + '.'
    else:
      prefix = self.module.GetNamespacePrefix()
    return '%s%s' % (prefix, self.mojom_name)

  def _tuple(self):
    return (self.mojom_name, self.native_only, self.fields, self.constants,
            self.attributes)

  def __eq__(self, rhs):
    return isinstance(rhs, Struct) and self._tuple() == rhs._tuple()

  def __lt__(self, rhs):
    if not isinstance(self, type(rhs)):
      return str(type(self)) < str(type(rhs))

    return self._tuple() < rhs._tuple()

  def __hash__(self):
    return id(self)


class Union(ReferenceKind):
  """A union of several kinds.

  Attributes:
    mojom_name: {str} The name of the union type as defined in mojom.
    name: {str} The stylized name.
    fields: {List[UnionField]} The members of the union.
    attributes: {dict} Additional information about the union, such as
        which Java class name to use to represent it in the generated
        bindings.
  """
  Kind.AddSharedProperty('mojom_name')
  Kind.AddSharedProperty('name')
  Kind.AddSharedProperty('fields')
  Kind.AddSharedProperty('attributes')
  Kind.AddSharedProperty('default_field')

  def __init__(self, mojom_name=None, module=None, attributes=None):
    if mojom_name is not None:
      spec = 'x:' + mojom_name
    else:
      spec = None
    ReferenceKind.__init__(self, spec, False, module)
    self.mojom_name = mojom_name
    self.name = None
    self.fields = []
    self.attributes = attributes
    self.default_field = None

  def Repr(self, as_ref=True):
    if as_ref:
      return '<%s spec=%r is_nullable=%r fields=%s>' % (
          self.__class__.__name__, self.spec, self.is_nullable, Repr(
              self.fields))
    return GenericRepr(self, {'fields': True, 'is_nullable': False})

  def AddField(self, mojom_name, kind, ordinal=None, attributes=None):
    field = UnionField(mojom_name, kind, ordinal, None, attributes)
    self.fields.append(field)
    return field

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeUnion(self.mojom_name)
    for field in self.fields:
      field.Stylize(stylizer)

  def IsBackwardCompatible(self, rhs, checker):
    """This union is backward-compatible with rhs (older_union) if and only if
    all of the following conditions hold:
      - Any newly added field is tagged with a [MinVersion] attribute specifying
        a version number greater than all previously used [MinVersion]
        attributes within the union.
      - All fields present in rhs remain present in the new union,
        with the same ordinal value, same optional or non-optional status,
        same (or backward-compatible) type, and where applicable, the same
        [MinVersion] attribute value.
    """

    def buildOrdinalFieldMap(union):
      fields_by_ordinal = {}
      for field in union.fields:
        if field.ordinal in fields_by_ordinal:
          raise Exception('Multiple fields with ordinal %s in union %s.' %
                          (field.ordinal, union.mojom_name))
        fields_by_ordinal[field.ordinal] = field
      return fields_by_ordinal

    new_fields = buildOrdinalFieldMap(self)
    old_fields = buildOrdinalFieldMap(rhs)
    if len(new_fields) < len(old_fields):
      # At least one field was removed, which is not OK.
      return False

    max_old_min_version = 0
    for ordinal, old_field in old_fields.items():
      new_field = new_fields.get(ordinal)
      if not new_field:
        # A field was removed, which is not OK.
        return False
      if not _IsFieldBackwardCompatible(new_field, old_field, checker):
        # An field changed its type or MinVersion, which is not OK.
        return False
      old_min_version = old_field.min_version or 0
      if old_min_version > max_old_min_version:
        max_old_min_version = old_min_version

    new_ordinals = set(new_fields.keys()) - set(old_fields.keys())
    for ordinal in new_ordinals:
      if (new_fields[ordinal].min_version or 0) <= max_old_min_version:
        # New fields must use a MinVersion greater than any old fields.
        return False

    return True

  @property
  def extensible(self):
    return self.attributes.get(ATTRIBUTE_EXTENSIBLE, False) \
        if self.attributes else False

  @property
  def stable(self):
    return self.attributes.get(ATTRIBUTE_STABLE, False) \
        if self.attributes else False

  @property
  def qualified_name(self):
    if self.parent_kind:
      prefix = self.parent_kind.qualified_name + '.'
    else:
      prefix = self.module.GetNamespacePrefix()
    return '%s%s' % (prefix, self.mojom_name)

  def _tuple(self):
    return (self.mojom_name, self.fields, self.attributes)

  def __eq__(self, rhs):
    return isinstance(rhs, Union) and self._tuple() == rhs._tuple()

  def __lt__(self, rhs):
    if not isinstance(self, type(rhs)):
      return str(type(self)) < str(type(rhs))

    return self._tuple() < rhs._tuple()

  def __hash__(self):
    return id(self)


class Array(ReferenceKind):
  """An array.

  Attributes:
    kind: {Kind} The type of the elements. May be None.
    length: The number of elements. None if unknown.
  """

  Kind.AddSharedProperty('kind')
  Kind.AddSharedProperty('length')

  def __init__(self, kind=None, length=None):
    if kind is not None:
      if length is not None:
        spec = 'a%d:%s' % (length, kind.spec)
      else:
        spec = 'a:%s' % kind.spec

      ReferenceKind.__init__(self, spec)
    else:
      ReferenceKind.__init__(self)
    self.kind = kind
    self.length = length

  def Repr(self, as_ref=True):
    if as_ref:
      return '<%s spec=%r is_nullable=%r kind=%s length=%r>' % (
          self.__class__.__name__, self.spec, self.is_nullable, Repr(
              self.kind), self.length)
    return GenericRepr(self, {
        'kind': True,
        'length': False,
        'is_nullable': False
    })

  def __eq__(self, rhs):
    return (isinstance(rhs, Array)
            and (self.kind, self.length) == (rhs.kind, rhs.length))

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return (isinstance(rhs, Array) and self.length == rhs.length
            and checker.IsBackwardCompatible(self.kind, rhs.kind))


class Map(ReferenceKind):
  """A map.

  Attributes:
    key_kind: {Kind} The type of the keys. May be None.
    value_kind: {Kind} The type of the elements. May be None.
  """
  Kind.AddSharedProperty('key_kind')
  Kind.AddSharedProperty('value_kind')

  def __init__(self, key_kind=None, value_kind=None):
    if (key_kind is not None and value_kind is not None):
      ReferenceKind.__init__(
          self, 'm[' + key_kind.spec + '][' + value_kind.spec + ']')
      if IsNullableKind(key_kind):
        raise Exception("Nullable kinds cannot be keys in maps.")
      if IsAnyHandleKind(key_kind):
        raise Exception("Handles cannot be keys in maps.")
      if IsAnyInterfaceKind(key_kind):
        raise Exception("Interfaces cannot be keys in maps.")
      if IsArrayKind(key_kind):
        raise Exception("Arrays cannot be keys in maps.")
    else:
      ReferenceKind.__init__(self)

    self.key_kind = key_kind
    self.value_kind = value_kind

  def Repr(self, as_ref=True):
    if as_ref:
      return '<%s spec=%r is_nullable=%r key_kind=%s value_kind=%s>' % (
          self.__class__.__name__, self.spec, self.is_nullable,
          Repr(self.key_kind), Repr(self.value_kind))
    return GenericRepr(self, {'key_kind': True, 'value_kind': True})

  def __eq__(self, rhs):
    return (isinstance(rhs, Map) and
            (self.key_kind, self.value_kind) == (rhs.key_kind, rhs.value_kind))

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return (isinstance(rhs, Map)
            and checker.IsBackwardCompatible(self.key_kind, rhs.key_kind)
            and checker.IsBackwardCompatible(self.value_kind, rhs.value_kind))


class PendingRemote(ReferenceKind):
  Kind.AddSharedProperty('kind')

  def __init__(self, kind=None):
    if kind is not None:
      if not isinstance(kind, Interface):
        raise Exception(
            'pending_remote<T> requires T to be an interface type. Got %r' %
            kind.spec)
      ReferenceKind.__init__(self, 'rmt:' + kind.spec)
    else:
      ReferenceKind.__init__(self)
    self.kind = kind

  def __eq__(self, rhs):
    return isinstance(rhs, PendingRemote) and self.kind == rhs.kind

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return (isinstance(rhs, PendingRemote)
            and checker.IsBackwardCompatible(self.kind, rhs.kind))


class PendingReceiver(ReferenceKind):
  Kind.AddSharedProperty('kind')

  def __init__(self, kind=None):
    if kind is not None:
      if not isinstance(kind, Interface):
        raise Exception(
            'pending_receiver<T> requires T to be an interface type. Got %r' %
            kind.spec)
      ReferenceKind.__init__(self, 'rcv:' + kind.spec)
    else:
      ReferenceKind.__init__(self)
    self.kind = kind

  def __eq__(self, rhs):
    return isinstance(rhs, PendingReceiver) and self.kind == rhs.kind

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return isinstance(rhs, PendingReceiver) and checker.IsBackwardCompatible(
        self.kind, rhs.kind)


class PendingAssociatedRemote(ReferenceKind):
  Kind.AddSharedProperty('kind')

  def __init__(self, kind=None):
    if kind is not None:
      if not isinstance(kind, Interface):
        raise Exception(
            'pending_associated_remote<T> requires T to be an interface ' +
            'type. Got %r' % kind.spec)
      ReferenceKind.__init__(self, 'rma:' + kind.spec)
    else:
      ReferenceKind.__init__(self)
    self.kind = kind

  def __eq__(self, rhs):
    return isinstance(rhs, PendingAssociatedRemote) and self.kind == rhs.kind

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return isinstance(rhs,
                      PendingAssociatedRemote) and checker.IsBackwardCompatible(
                          self.kind, rhs.kind)


class PendingAssociatedReceiver(ReferenceKind):
  Kind.AddSharedProperty('kind')

  def __init__(self, kind=None):
    if kind is not None:
      if not isinstance(kind, Interface):
        raise Exception(
            'pending_associated_receiver<T> requires T to be an interface' +
            'type. Got %r' % kind.spec)
      ReferenceKind.__init__(self, 'rca:' + kind.spec)
    else:
      ReferenceKind.__init__(self)
    self.kind = kind

  def __eq__(self, rhs):
    return isinstance(rhs, PendingAssociatedReceiver) and self.kind == rhs.kind

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return isinstance(
        rhs, PendingAssociatedReceiver) and checker.IsBackwardCompatible(
            self.kind, rhs.kind)


class InterfaceRequest(ReferenceKind):
  Kind.AddSharedProperty('kind')

  def __init__(self, kind=None):
    if kind is not None:
      if not isinstance(kind, Interface):
        raise Exception(
            "Interface request requires %r to be an interface." % kind.spec)
      ReferenceKind.__init__(self, 'r:' + kind.spec)
    else:
      ReferenceKind.__init__(self)
    self.kind = kind

  def __eq__(self, rhs):
    return isinstance(rhs, InterfaceRequest) and self.kind == rhs.kind

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return isinstance(rhs, InterfaceRequest) and checker.IsBackwardCompatible(
        self.kind, rhs.kind)


class AssociatedInterfaceRequest(ReferenceKind):
  Kind.AddSharedProperty('kind')

  def __init__(self, kind=None):
    if kind is not None:
      if not isinstance(kind, InterfaceRequest):
        raise Exception(
            "Associated interface request requires %r to be an interface "
            "request." % kind.spec)
      assert not kind.is_nullable
      ReferenceKind.__init__(self, 'asso:' + kind.spec)
    else:
      ReferenceKind.__init__(self)
    self.kind = kind.kind if kind is not None else None

  def __eq__(self, rhs):
    return isinstance(rhs, AssociatedInterfaceRequest) and self.kind == rhs.kind

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return isinstance(
        rhs, AssociatedInterfaceRequest) and checker.IsBackwardCompatible(
            self.kind, rhs.kind)


class Parameter:
  def __init__(self,
               mojom_name=None,
               kind=None,
               ordinal=None,
               default=None,
               attributes=None):
    self.mojom_name = mojom_name
    self.name = None
    self.ordinal = ordinal
    self.kind = kind
    self.default = default
    self.attributes = attributes

  def Repr(self, as_ref=True):
    # pylint: disable=unused-argument
    return '<%s mojom_name=%r kind=%s>' % (
        self.__class__.__name__, self.mojom_name, self.kind.Repr(as_ref=True))

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeParameter(self.mojom_name)

  @property
  def min_version(self):
    return self.attributes.get(ATTRIBUTE_MIN_VERSION) \
        if self.attributes else None

  def __eq__(self, rhs):
    return (isinstance(rhs, Parameter)
            and (self.mojom_name, self.ordinal, self.kind, self.default,
                 self.attributes) == (rhs.mojom_name, rhs.ordinal, rhs.kind,
                                      rhs.default, rhs.attributes))


class Method:
  def __init__(self, interface, mojom_name, ordinal=None, attributes=None):
    self.interface = interface
    self.mojom_name = mojom_name
    self.name = None
    self.explicit_ordinal = ordinal
    self.ordinal = ordinal
    self.parameters = []
    self.param_struct = None
    self.response_parameters = None
    self.response_param_struct = None
    self.attributes = attributes

  def Repr(self, as_ref=True):
    if as_ref:
      return '<%s mojom_name=%r>' % (self.__class__.__name__, self.mojom_name)
    return GenericRepr(self, {
        'mojom_name': False,
        'parameters': True,
        'response_parameters': True
    })

  def AddParameter(self,
                   mojom_name,
                   kind,
                   ordinal=None,
                   default=None,
                   attributes=None):
    parameter = Parameter(mojom_name, kind, ordinal, default, attributes)
    self.parameters.append(parameter)
    return parameter

  def AddResponseParameter(self,
                           mojom_name,
                           kind,
                           ordinal=None,
                           default=None,
                           attributes=None):
    if self.response_parameters == None:
      self.response_parameters = []
    parameter = Parameter(mojom_name, kind, ordinal, default, attributes)
    self.response_parameters.append(parameter)
    return parameter

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeMethod(self.mojom_name)
    for param in self.parameters:
      param.Stylize(stylizer)
    if self.response_parameters is not None:
      for param in self.response_parameters:
        param.Stylize(stylizer)

    if self.param_struct:
      self.param_struct.Stylize(stylizer)
    if self.response_param_struct:
      self.response_param_struct.Stylize(stylizer)

  @property
  def min_version(self):
    return self.attributes.get(ATTRIBUTE_MIN_VERSION) \
        if self.attributes else None

  @property
  def sync(self):
    return self.attributes.get(ATTRIBUTE_SYNC) \
        if self.attributes else None

  @property
  def allow_interrupt(self):
    return not self.attributes.get(ATTRIBUTE_NO_INTERRUPT) \
        if self.attributes else True

  @property
  def unlimited_message_size(self):
    return self.attributes.get(ATTRIBUTE_UNLIMITED_SIZE) \
        if self.attributes else False

  @property
  def allowed_context(self):
    return self.attributes.get(ATTRIBUTE_ALLOWED_CONTEXT) \
        if self.attributes else None

  @property
  def supports_urgent(self):
    return self.attributes.get(ATTRIBUTE_SUPPORTS_URGENT) \
        if self.attributes else None

  @property
  def runtime_feature(self):
    if not self.attributes:
      return None
    runtime_feature = self.attributes.get(ATTRIBUTE_RUNTIME_FEATURE, None)
    if runtime_feature is None:
      return None
    if not isinstance(runtime_feature, Feature):
      raise Exception("RuntimeFeature attribute on %s must be a feature." %
                      self.name)
    return runtime_feature

  def _tuple(self):
    return (self.mojom_name, self.ordinal, self.parameters,
            self.response_parameters, self.attributes)

  def __eq__(self, rhs):
    return isinstance(rhs, Method) and self._tuple() == rhs._tuple()

  def __lt__(self, rhs):
    if not isinstance(self, type(rhs)):
      return str(type(self)) < str(type(rhs))

    return self._tuple() < rhs._tuple()


class Interface(ReferenceKind):
  Kind.AddSharedProperty('mojom_name')
  Kind.AddSharedProperty('name')
  Kind.AddSharedProperty('methods')
  Kind.AddSharedProperty('enums')
  Kind.AddSharedProperty('constants')
  Kind.AddSharedProperty('attributes')

  def __init__(self, mojom_name=None, module=None, attributes=None):
    if mojom_name is not None:
      spec = 'x:' + mojom_name
    else:
      spec = None
    ReferenceKind.__init__(self, spec, False, module)
    self.mojom_name = mojom_name
    self.name = None
    self.methods = []
    self.enums = []
    self.constants = []
    self.attributes = attributes

  def Repr(self, as_ref=True):
    if as_ref:
      return '<%s mojom_name=%r>' % (self.__class__.__name__, self.mojom_name)
    return GenericRepr(self, {
        'mojom_name': False,
        'attributes': False,
        'methods': False
    })

  def AddMethod(self, mojom_name, ordinal=None, attributes=None):
    method = Method(self, mojom_name, ordinal, attributes)
    self.methods.append(method)
    return method

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeInterface(self.mojom_name)
    for method in self.methods:
      method.Stylize(stylizer)
    for enum in self.enums:
      enum.Stylize(stylizer)
    for constant in self.constants:
      constant.Stylize(stylizer)

  def IsBackwardCompatible(self, rhs, checker):
    """This interface is backward-compatible with rhs (older_interface) if and
    only if all of the following conditions hold:
      - All defined methods in rhs (when identified by ordinal) have
        backward-compatible definitions in this interface. For each method this
        means:
          - The parameter list is backward-compatible, according to backward-
            compatibility rules for structs, where each parameter is essentially
            a struct field.
          - If the old method definition does not specify a reply message, the
            new method definition must not specify a reply message.
          - If the old method definition specifies a reply message, the new
            method definition must also specify a reply message with a parameter
            list that is backward-compatible according to backward-compatibility
            rules for structs.
      - All newly introduced methods in this interface have a [MinVersion]
        attribute specifying a version greater than any method in
        rhs.
    """

    def buildOrdinalMethodMap(interface):
      methods_by_ordinal = {}
      for method in interface.methods:
        if method.ordinal in methods_by_ordinal:
          raise Exception('Multiple methods with ordinal %s in interface %s.' %
                          (method.ordinal, interface.mojom_name))
        methods_by_ordinal[method.ordinal] = method
      return methods_by_ordinal

    new_methods = buildOrdinalMethodMap(self)
    old_methods = buildOrdinalMethodMap(rhs)
    max_old_min_version = 0
    for ordinal, old_method in old_methods.items():
      new_method = new_methods.get(ordinal)
      if not new_method:
        # A method was removed, which is not OK.
        return False

      if not checker.IsBackwardCompatible(new_method.param_struct,
                                          old_method.param_struct):
        # The parameter list is not backward-compatible, which is not OK.
        return False

      if old_method.response_param_struct is None:
        if new_method.response_param_struct is not None:
          # A reply was added to a message which didn't have one before, and
          # this is not OK.
          return False
      else:
        if new_method.response_param_struct is None:
          # A reply was removed from a message, which is not OK.
          return False
        if not checker.IsBackwardCompatible(new_method.response_param_struct,
                                            old_method.response_param_struct):
          # The new message's reply is not backward-compatible with the old
          # message's reply, which is not OK.
          return False

      if (old_method.min_version or 0) > max_old_min_version:
        max_old_min_version = old_method.min_version

    # All the old methods are compatible with their new counterparts. Now verify
    # that newly added methods are properly versioned.
    new_ordinals = set(new_methods.keys()) - set(old_methods.keys())
    for ordinal in new_ordinals:
      new_method = new_methods[ordinal]
      if (new_method.min_version or 0) <= max_old_min_version:
        # A method was added to an existing version, which is not OK.
        return False

    return True

  @property
  def service_sandbox(self):
    if not self.attributes:
      return None
    service_sandbox = self.attributes.get(ATTRIBUTE_SERVICE_SANDBOX, None)
    if service_sandbox is None:
      return None
    # Constants are only allowed to refer to an enum here, so replace.
    if isinstance(service_sandbox, Constant):
      service_sandbox = service_sandbox.value
    if not isinstance(service_sandbox, EnumValue):
      raise Exception("ServiceSandbox attribute on %s must be an enum value." %
                      self.module.name)
    return service_sandbox

  @property
  def runtime_feature(self):
    if not self.attributes:
      return None
    runtime_feature = self.attributes.get(ATTRIBUTE_RUNTIME_FEATURE, None)
    if runtime_feature is None:
      return None
    if not isinstance(runtime_feature, Feature):
      raise Exception("RuntimeFeature attribute on %s must be a feature." %
                      self.name)
    return runtime_feature

  @property
  def require_context(self):
    if not self.attributes:
      return None
    return self.attributes.get(ATTRIBUTE_REQUIRE_CONTEXT, None)

  @property
  def stable(self):
    return self.attributes.get(ATTRIBUTE_STABLE, False) \
        if self.attributes else False

  @property
  def qualified_name(self):
    if self.parent_kind:
      prefix = self.parent_kind.qualified_name + '.'
    else:
      prefix = self.module.GetNamespacePrefix()
    return '%s%s' % (prefix, self.mojom_name)

  def _tuple(self):
    return (self.mojom_name, self.methods, self.enums, self.constants,
            self.attributes)

  def __eq__(self, rhs):
    return isinstance(rhs, Interface) and self._tuple() == rhs._tuple()

  def __lt__(self, rhs):
    if not isinstance(self, type(rhs)):
      return str(type(self)) < str(type(rhs))

    return self._tuple() < rhs._tuple()

  @property
  def uuid(self):
    uuid_str = self.attributes.get(ATTRIBUTE_UUID) if self.attributes else None
    if uuid_str is None:
      return None

    try:
      u = UUID(uuid_str)
    except:
      raise ValueError('Invalid format for Uuid attribute on interface {}. '
                       'Expected standard RFC 4122 string representation of '
                       'a UUID.'.format(self.mojom_name))
    return (int(u.hex[:16], 16), int(u.hex[16:], 16))

  def __hash__(self):
    return id(self)


class AssociatedInterface(ReferenceKind):
  Kind.AddSharedProperty('kind')

  def __init__(self, kind=None):
    if kind is not None:
      if not isinstance(kind, Interface):
        raise Exception(
            "Associated interface requires %r to be an interface." % kind.spec)
      assert not kind.is_nullable
      ReferenceKind.__init__(self, 'asso:' + kind.spec)
    else:
      ReferenceKind.__init__(self)
    self.kind = kind

  def __eq__(self, rhs):
    return isinstance(rhs, AssociatedInterface) and self.kind == rhs.kind

  def __hash__(self):
    return id(self)

  def IsBackwardCompatible(self, rhs, checker):
    return isinstance(rhs,
                      AssociatedInterface) and checker.IsBackwardCompatible(
                          self.kind, rhs.kind)


class EnumField:
  def __init__(self,
               mojom_name=None,
               value=None,
               attributes=None,
               numeric_value=None):
    self.mojom_name = mojom_name
    self.name = None
    self.value = value
    self.attributes = attributes
    self.numeric_value = numeric_value

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeEnumField(self.mojom_name)

  @property
  def default(self):
    return self.attributes.get(ATTRIBUTE_DEFAULT, False) \
        if self.attributes else False

  @property
  def min_version(self):
    return self.attributes.get(ATTRIBUTE_MIN_VERSION) \
        if self.attributes else None

  def __eq__(self, rhs):
    return (isinstance(rhs, EnumField)
            and (self.mojom_name, self.value, self.attributes,
                 self.numeric_value) == (rhs.mojom_name, rhs.value,
                                         rhs.attributes, rhs.numeric_value))


class Enum(ValueKind):
  Kind.AddSharedProperty('mojom_name')
  Kind.AddSharedProperty('name')
  Kind.AddSharedProperty('native_only')
  Kind.AddSharedProperty('fields')
  Kind.AddSharedProperty('attributes')
  Kind.AddSharedProperty('min_value')
  Kind.AddSharedProperty('max_value')
  Kind.AddSharedProperty('default_field')

  def __init__(self, mojom_name=None, module=None, attributes=None):
    if mojom_name is not None:
      spec = 'x:' + mojom_name
    else:
      spec = None
    ValueKind.__init__(self, spec, False, module)
    self.mojom_name = mojom_name
    self.name = None
    self.native_only = False
    self.fields = []
    self.attributes = attributes
    self.min_value = None
    self.max_value = None
    self.default_field = None

  def Repr(self, as_ref=True):
    if as_ref:
      return '<%s mojom_name=%r>' % (self.__class__.__name__, self.mojom_name)
    return GenericRepr(self, {'mojom_name': False, 'fields': False})

  def Stylize(self, stylizer):
    self.name = stylizer.StylizeEnum(self.mojom_name)
    for field in self.fields:
      field.Stylize(stylizer)

  @property
  def extensible(self):
    return self.attributes.get(ATTRIBUTE_EXTENSIBLE, False) \
        if self.attributes else False

  @property
  def stable(self):
    return self.attributes.get(ATTRIBUTE_STABLE, False) \
        if self.attributes else False

  @property
  def qualified_name(self):
    if self.parent_kind:
      prefix = self.parent_kind.qualified_name + '.'
    else:
      prefix = self.module.GetNamespacePrefix()
    return '%s%s' % (prefix, self.mojom_name)

  # pylint: disable=unused-argument
  def IsBackwardCompatible(self, rhs, checker):
    """This enum is backward-compatible with rhs (older_enum) if and only if one
    of the following conditions holds:
        - Neither enum is [Extensible] and both have the exact same set of valid
          numeric values. Field names and aliases for the same numeric value do
          not affect compatibility.
        - rhs is [Extensible], and for every version defined by
          rhs, this enum has the exact same set of valid numeric values.
    """

    def buildVersionFieldMap(enum):
      fields_by_min_version = {}
      for field in enum.fields:
        if field.min_version not in fields_by_min_version:
          fields_by_min_version[field.min_version] = set()
        fields_by_min_version[field.min_version].add(field.numeric_value)
      return fields_by_min_version

    old_fields = buildVersionFieldMap(rhs)
    new_fields = buildVersionFieldMap(self)

    if new_fields.keys() != old_fields.keys() and not rhs.extensible:
      raise Exception("Non-extensible enum cannot be modified")

    for min_version, valid_values in old_fields.items():
      if min_version not in new_fields:
        raise Exception('New values added to an extensible enum '
                        'do not specify MinVersion: %s' % new_fields)

      if (new_fields[min_version] != valid_values):
        if (len(new_fields[min_version]) < len(valid_values)):
          raise Exception('Removing values for an existing MinVersion %s '
                          'is not allowed' % min_version)

        raise Exception(
            'New values don\'t match old values'
            'for an existing MinVersion %s,'
            ' please specify MinVersion equal to "Next version" '
            'in the enum description'
            ' for the following values:\n%s' %
            (min_version, new_fields[min_version].difference(valid_values)))
    return True

  def _tuple(self):
    return (self.mojom_name, self.native_only, self.fields, self.attributes,
            self.min_value, self.max_value, self.default_field)

  def __eq__(self, rhs):
    return isinstance(rhs, Enum) and self._tuple() == rhs._tuple()

  def __lt__(self, rhs):
    if not isinstance(self, type(rhs)):
      return str(type(self)) < str(type(rhs))

    return self._tuple() < rhs._tuple()

  def __hash__(self):
    return id(self)


class Module:
  def __init__(self, path=None, mojom_namespace=None, attributes=None):
    self.path = path
    self.mojom_namespace = mojom_namespace
    self.namespace = None
    self.structs = []
    self.unions = []
    self.interfaces = []
    self.enums = []
    self.features = []
    self.constants = []
    self.kinds = OrderedDict()
    self.attributes = attributes
    self.imports = []
    self.imported_kinds = OrderedDict()
    self.metadata = OrderedDict()

  def __repr__(self):
    # Gives us a decent __repr__ for modules.
    return self.Repr()

  def __eq__(self, rhs):
    return (isinstance(rhs, Module)
            and (self.path, self.attributes, self.mojom_namespace, self.imports,
                 self.constants, self.enums, self.structs, self.unions,
                 self.interfaces, self.features)
            == (rhs.path, rhs.attributes, rhs.mojom_namespace, rhs.imports,
                rhs.constants, rhs.enums, rhs.structs, rhs.unions,
                rhs.interfaces, rhs.features))

  def __hash__(self):
    return id(self)

  def Repr(self, as_ref=True):
    if as_ref:
      return '<%s path=%r mojom_namespace=%r>' % (
          self.__class__.__name__, self.path, self.mojom_namespace)
    return GenericRepr(
        self, {
            'path': False,
            'mojom_namespace': False,
            'attributes': False,
            'structs': False,
            'interfaces': False,
            'unions': False,
            'features': False,
        })

  def GetNamespacePrefix(self):
    return '%s.' % self.mojom_namespace if self.mojom_namespace else ''

  def AddInterface(self, mojom_name, attributes=None):
    interface = Interface(mojom_name, self, attributes)
    self.interfaces.append(interface)
    return interface

  def AddStruct(self, mojom_name, attributes=None):
    struct = Struct(mojom_name, self, attributes)
    self.structs.append(struct)
    return struct

  def AddUnion(self, mojom_name, attributes=None):
    union = Union(mojom_name, self, attributes)
    self.unions.append(union)
    return union

  def AddFeature(self, mojom_name, attributes=None):
    feature = Feature(mojom_name, self, attributes)
    self.features.append(feature)
    return feature

  def Stylize(self, stylizer):
    self.namespace = stylizer.StylizeModule(self.mojom_namespace)
    for struct in self.structs:
      struct.Stylize(stylizer)
    for union in self.unions:
      union.Stylize(stylizer)
    for interface in self.interfaces:
      interface.Stylize(stylizer)
    for enum in self.enums:
      enum.Stylize(stylizer)
    for constant in self.constants:
      constant.Stylize(stylizer)
    for feature in self.features:
      feature.Stylize(stylizer)

    for imported_module in self.imports:
      imported_module.Stylize(stylizer)

  def Dump(self, f):
    pickle.dump(self, f)

  @classmethod
  def Load(cls, f):
    result = pickle.load(f)
    assert isinstance(result, Module)
    return result


def IsBoolKind(kind):
  return kind.spec == BOOL.spec or kind.spec == NULLABLE_BOOL.spec


def IsFloatKind(kind):
  return kind.spec == FLOAT.spec or kind.spec == NULLABLE_FLOAT.spec


def IsDoubleKind(kind):
  return kind.spec == DOUBLE.spec or kind.spec == NULLABLE_DOUBLE.spec


def IsIntegralKind(kind):
  return (kind.spec == BOOL.spec or kind.spec == INT8.spec
          or kind.spec == INT16.spec or kind.spec == INT32.spec
          or kind.spec == INT64.spec or kind.spec == UINT8.spec
          or kind.spec == UINT16.spec or kind.spec == UINT32.spec
          or kind.spec == UINT64.spec or kind.spec == NULLABLE_BOOL.spec
          or kind.spec == NULLABLE_INT8.spec or kind.spec == NULLABLE_INT16.spec
          or kind.spec == NULLABLE_INT32.spec
          or kind.spec == NULLABLE_INT64.spec
          or kind.spec == NULLABLE_UINT8.spec
          or kind.spec == NULLABLE_UINT16.spec
          or kind.spec == NULLABLE_UINT32.spec
          or kind.spec == NULLABLE_UINT64.spec)


def IsStringKind(kind):
  return kind.spec == STRING.spec or kind.spec == NULLABLE_STRING.spec


def IsGenericHandleKind(kind):
  return kind.spec == HANDLE.spec or kind.spec == NULLABLE_HANDLE.spec


def IsDataPipeConsumerKind(kind):
  return kind.spec == DCPIPE.spec or kind.spec == NULLABLE_DCPIPE.spec


def IsDataPipeProducerKind(kind):
  return kind.spec == DPPIPE.spec or kind.spec == NULLABLE_DPPIPE.spec


def IsMessagePipeKind(kind):
  return kind.spec == MSGPIPE.spec or kind.spec == NULLABLE_MSGPIPE.spec


def IsSharedBufferKind(kind):
  return (kind.spec == SHAREDBUFFER.spec
          or kind.spec == NULLABLE_SHAREDBUFFER.spec)


def IsPlatformHandleKind(kind):
  return (kind.spec == PLATFORMHANDLE.spec
          or kind.spec == NULLABLE_PLATFORMHANDLE.spec)


def IsStructKind(kind):
  return isinstance(kind, Struct)


def IsUnionKind(kind):
  return isinstance(kind, Union)


def IsArrayKind(kind):
  return isinstance(kind, Array)


def IsFeatureKind(kind):
  return isinstance(kind, Feature)


def IsInterfaceKind(kind):
  return isinstance(kind, Interface)


def IsAssociatedInterfaceKind(kind):
  return isinstance(kind, AssociatedInterface)


def IsInterfaceRequestKind(kind):
  return isinstance(kind, InterfaceRequest)


def IsAssociatedInterfaceRequestKind(kind):
  return isinstance(kind, AssociatedInterfaceRequest)


def IsPendingRemoteKind(kind):
  return isinstance(kind, PendingRemote)


def IsPendingReceiverKind(kind):
  return isinstance(kind, PendingReceiver)


def IsPendingAssociatedRemoteKind(kind):
  return isinstance(kind, PendingAssociatedRemote)


def IsPendingAssociatedReceiverKind(kind):
  return isinstance(kind, PendingAssociatedReceiver)


def IsEnumKind(kind):
  return isinstance(kind, Enum)


def IsValueKind(kind):
  return isinstance(kind, ValueKind)


def IsReferenceKind(kind):
  return isinstance(kind, ReferenceKind)


def IsNullableKind(kind):
  return kind.is_nullable


def IsMapKind(kind):
  return isinstance(kind, Map)


def IsObjectKind(kind):
  return IsPointerKind(kind) or IsUnionKind(kind)


def IsPointerKind(kind):
  return (IsStructKind(kind) or IsArrayKind(kind) or IsStringKind(kind)
          or IsMapKind(kind))


# Please note that it doesn't include any interface kind.
def IsAnyHandleKind(kind):
  return (IsGenericHandleKind(kind) or IsDataPipeConsumerKind(kind)
          or IsDataPipeProducerKind(kind) or IsMessagePipeKind(kind)
          or IsSharedBufferKind(kind) or IsPlatformHandleKind(kind))


def IsAnyInterfaceKind(kind):
  return (IsInterfaceKind(kind) or IsInterfaceRequestKind(kind)
          or IsAssociatedKind(kind) or IsPendingRemoteKind(kind)
          or IsPendingReceiverKind(kind))


def IsAnyHandleOrInterfaceKind(kind):
  return IsAnyHandleKind(kind) or IsAnyInterfaceKind(kind)


def IsAssociatedKind(kind):
  return (IsAssociatedInterfaceKind(kind)
          or IsAssociatedInterfaceRequestKind(kind)
          or IsPendingAssociatedRemoteKind(kind)
          or IsPendingAssociatedReceiverKind(kind))


def HasCallbacks(interface):
  for method in interface.methods:
    if method.response_parameters != None:
      return True
  return False


# Finds out whether an interface passes associated interfaces and associated
# interface requests.
def PassesAssociatedKinds(interface):
  visited_kinds = set()
  for method in interface.methods:
    if MethodPassesAssociatedKinds(method, visited_kinds):
      return True
  return False


def _AnyMethodParameterRecursive(method, predicate, visited_kinds=None):
  def _HasProperty(kind):
    if kind in visited_kinds:
      # No need to examine the kind again.
      return False
    visited_kinds.add(kind)
    if predicate(kind):
      return True
    if IsArrayKind(kind):
      return _HasProperty(kind.kind)
    if IsStructKind(kind) or IsUnionKind(kind):
      for field in kind.fields:
        if _HasProperty(field.kind):
          return True
    if IsMapKind(kind):
      if _HasProperty(kind.key_kind) or _HasProperty(kind.value_kind):
        return True
    return False

  if visited_kinds is None:
    visited_kinds = set()

  for param in method.parameters:
    if _HasProperty(param.kind):
      return True
  if method.response_parameters != None:
    for param in method.response_parameters:
      if _HasProperty(param.kind):
        return True
  return False


# Finds out whether a method passes associated interfaces and associated
# interface requests.
def MethodPassesAssociatedKinds(method, visited_kinds=None):
  return _AnyMethodParameterRecursive(
      method, IsAssociatedKind, visited_kinds=visited_kinds)


# Determines whether a method passes interfaces.
def MethodPassesInterfaces(method):
  return _AnyMethodParameterRecursive(method, IsInterfaceKind)


def GetSyncMethodOrdinals(interface):
  return [method.ordinal for method in interface.methods if method.sync]


def HasUninterruptableMethods(interface):
  for method in interface.methods:
    if not method.allow_interrupt:
      return True
  return False


def ContainsHandlesOrInterfaces(kind):
  """Check if the kind contains any handles.

  This check is recursive so it checks all struct fields, containers elements,
  etc.

  Args:
    struct: {Kind} The kind to check.

  Returns:
    {bool}: True if the kind contains handles.
  """
  # We remember the types we already checked to avoid infinite recursion when
  # checking recursive (or mutually recursive) types:
  checked = set()

  def Check(kind):
    if kind.spec in checked:
      return False
    checked.add(kind.spec)
    if IsStructKind(kind):
      return any(Check(field.kind) for field in kind.fields)
    if IsUnionKind(kind):
      return any(Check(field.kind) for field in kind.fields)
    if IsAnyHandleKind(kind):
      return True
    if IsAnyInterfaceKind(kind):
      return True
    if IsArrayKind(kind):
      return Check(kind.kind)
    if IsMapKind(kind):
      return Check(kind.key_kind) or Check(kind.value_kind)
    return False

  return Check(kind)


def ContainsNativeTypes(kind):
  """Check if the kind contains any native type (struct or enum).

  This check is recursive so it checks all struct fields, scoped interface
  enums, etc.

  Args:
    struct: {Kind} The kind to check.

  Returns:
    {bool}: True if the kind contains native types.
  """
  # We remember the types we already checked to avoid infinite recursion when
  # checking recursive (or mutually recursive) types:
  checked = set()

  def Check(kind):
    if kind.spec in checked:
      return False
    checked.add(kind.spec)
    if IsEnumKind(kind):
      return kind.native_only
    if IsStructKind(kind):
      if kind.native_only:
        return True
      if any(enum.native_only for enum in kind.enums):
        return True
      return any(Check(field.kind) for field in kind.fields)
    if IsUnionKind(kind):
      return any(Check(field.kind) for field in kind.fields)
    if IsInterfaceKind(kind):
      return any(enum.native_only for enum in kind.enums)
    if IsArrayKind(kind):
      return Check(kind.kind)
    if IsMapKind(kind):
      return Check(kind.key_kind) or Check(kind.value_kind)
    return False

  return Check(kind)



# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2022, Tomi Valkeinen <tomi.valkeinen@ideasonboard.com>

from PyQt6 import QtCore, QtWidgets
from PyQt6.QtCore import Qt

import math
import os
import sys

os.environ['PYOPENGL_PLATFORM'] = 'egl'

from OpenGL.EGL.EXT.image_dma_buf_import import *
from OpenGL.EGL.KHR.image import *
from OpenGL.EGL.VERSION.EGL_1_0 import *
from OpenGL.EGL.VERSION.EGL_1_2 import *
from OpenGL.EGL.VERSION.EGL_1_3 import *

from OpenGL.GLES2.OES.EGL_image import *
from OpenGL.GLES2.OES.EGL_image_external import *
from OpenGL.GLES2.VERSION.GLES2_2_0 import *
from OpenGL.GLES3.VERSION.GLES3_3_0 import *

from OpenGL.GL import shaders

from gl_helpers import *


class EglState:
    def __init__(self):
        self.create_display()
        self.choose_config()
        self.create_context()
        self.check_extensions()

    def create_display(self):
        xdpy = getEGLNativeDisplay()
        dpy = eglGetDisplay(xdpy)
        self.display = dpy

    def choose_config(self):
        dpy = self.display

        major, minor = EGLint(), EGLint()

        b = eglInitialize(dpy, major, minor)
        assert(b)

        print('EGL {} {}'.format(
              eglQueryString(dpy, EGL_VENDOR).decode(),
              eglQueryString(dpy, EGL_VERSION).decode()))

        check_egl_extensions(dpy, ['EGL_EXT_image_dma_buf_import'])

        b = eglBindAPI(EGL_OPENGL_ES_API)
        assert(b)

        def print_config(dpy, cfg):

            def getconf(a):
                value = ctypes.c_long()
                eglGetConfigAttrib(dpy, cfg, a, value)
                return value.value

            print('EGL Config {}: color buf {}/{}/{}/{} = {}, depth {}, stencil {}, native visualid {}, native visualtype {}'.format(
                getconf(EGL_CONFIG_ID),
                getconf(EGL_ALPHA_SIZE),
                getconf(EGL_RED_SIZE),
                getconf(EGL_GREEN_SIZE),
                getconf(EGL_BLUE_SIZE),
                getconf(EGL_BUFFER_SIZE),
                getconf(EGL_DEPTH_SIZE),
                getconf(EGL_STENCIL_SIZE),
                getconf(EGL_NATIVE_VISUAL_ID),
                getconf(EGL_NATIVE_VISUAL_TYPE)))

        if False:
            num_configs = ctypes.c_long()
            eglGetConfigs(dpy, None, 0, num_configs)
            print('{} configs'.format(num_configs.value))

            configs = (EGLConfig * num_configs.value)()
            eglGetConfigs(dpy, configs, num_configs.value, num_configs)
            for config_id in configs:
                print_config(dpy, config_id)

        config_attribs = [
            EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
            EGL_RED_SIZE, 8,
            EGL_GREEN_SIZE, 8,
            EGL_BLUE_SIZE, 8,
            EGL_ALPHA_SIZE, 0,
            EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
            EGL_NONE,
        ]

        n = EGLint()
        configs = (EGLConfig * 1)()
        b = eglChooseConfig(dpy, config_attribs, configs, 1, n)
        assert(b and n.value == 1)
        config = configs[0]

        print('Chosen Config:')
        print_config(dpy, config)

        self.config = config

    def create_context(self):
        dpy = self.display

        context_attribs = [
            EGL_CONTEXT_CLIENT_VERSION, 2,
            EGL_NONE,
        ]

        context = eglCreateContext(dpy, self.config, EGL_NO_CONTEXT, context_attribs)
        assert(context)

        b = eglMakeCurrent(dpy, EGL_NO_SURFACE, EGL_NO_SURFACE, context)
        assert(b)

        self.context = context

    def check_extensions(self):
        check_gl_extensions(['GL_OES_EGL_image'])

        assert(eglCreateImageKHR)
        assert(eglDestroyImageKHR)
        assert(glEGLImageTargetTexture2DOES)


class QtRenderer:
    def __init__(self, state):
        self.state = state

    def setup(self):
        self.app = QtWidgets.QApplication([])

        window = MainWindow(self.state)
        window.show()

        self.window = window

    def run(self):
        camnotif = QtCore.QSocketNotifier(self.state.cm.event_fd, QtCore.QSocketNotifier.Type.Read)
        camnotif.activated.connect(lambda _: self.readcam())

        keynotif = QtCore.QSocketNotifier(sys.stdin.fileno(), QtCore.QSocketNotifier.Type.Read)
        keynotif.activated.connect(lambda _: self.readkey())

        print('Capturing...')

        self.app.exec()

        print('Exiting...')

    def readcam(self):
        running = self.state.event_handler()

        if not running:
            self.app.quit()

    def readkey(self):
        sys.stdin.readline()
        self.app.quit()

    def request_handler(self, ctx, req):
        self.window.handle_request(ctx, req)

    def cleanup(self):
        self.window.close()


class MainWindow(QtWidgets.QWidget):
    def __init__(self, state):
        super().__init__()

        self.setAttribute(Qt.WidgetAttribute.WA_PaintOnScreen)
        self.setAttribute(Qt.WidgetAttribute.WA_NativeWindow)

        self.state = state

        self.textures = {}
        self.reqqueue = {}
        self.current = {}

        for ctx in self.state.contexts:

            self.reqqueue[ctx.idx] = []
            self.current[ctx.idx] = []

            for stream in ctx.streams:
                self.textures[stream] = None

        num_tiles = len(self.textures)
        self.num_columns = math.ceil(math.sqrt(num_tiles))
        self.num_rows = math.ceil(num_tiles / self.num_columns)

        self.egl = EglState()

        self.surface = None

    def paintEngine(self):
        return None

    def create_surface(self):
        native_surface = c_void_p(self.winId().__int__())
        surface = eglCreateWindowSurface(self.egl.display, self.egl.config,
                                         native_surface, None)

        b = eglMakeCurrent(self.egl.display, self.surface, self.surface, self.egl.context)
        assert(b)

        self.surface = surface

    def init_gl(self):
        self.create_surface()

        vertShaderSrc = '''
            attribute vec2 aPosition;
            varying vec2 texcoord;

            void main()
            {
                gl_Position = vec4(aPosition * 2.0 - 1.0, 0.0, 1.0);
                texcoord.x = aPosition.x;
                texcoord.y = 1.0 - aPosition.y;
            }
        '''
        fragShaderSrc = '''
            #extension GL_OES_EGL_image_external : enable
            precision mediump float;
            varying vec2 texcoord;
            uniform samplerExternalOES texture;

            void main()
            {
                gl_FragColor = texture2D(texture, texcoord);
            }
        '''

        program = shaders.compileProgram(
            shaders.compileShader(vertShaderSrc, GL_VERTEX_SHADER),
            shaders.compileShader(fragShaderSrc, GL_FRAGMENT_SHADER)
        )

        glUseProgram(program)

        glClearColor(0.5, 0.8, 0.7, 1.0)

        vertPositions = [
            0.0, 0.0,
            1.0, 0.0,
            1.0, 1.0,
            0.0, 1.0
        ]

        inputAttrib = glGetAttribLocation(program, 'aPosition')
        glVertexAttribPointer(inputAttrib, 2, GL_FLOAT, GL_FALSE, 0, vertPositions)
        glEnableVertexAttribArray(inputAttrib)

    def create_texture(self, stream, fb):
        cfg = stream.configuration
        fmt = cfg.pixel_format.fourcc
        w = cfg.size.width
        h = cfg.size.height

        attribs = [
            EGL_WIDTH, w,
            EGL_HEIGHT, h,
            EGL_LINUX_DRM_FOURCC_EXT, fmt,
            EGL_DMA_BUF_PLANE0_FD_EXT, fb.planes[0].fd,
            EGL_DMA_BUF_PLANE0_OFFSET_EXT, 0,
            EGL_DMA_BUF_PLANE0_PITCH_EXT, cfg.stride,
            EGL_NONE,
        ]

        image = eglCreateImageKHR(self.egl.display,
                                  EGL_NO_CONTEXT,
                                  EGL_LINUX_DMA_BUF_EXT,
                                  None,
                                  attribs)
        assert(image)

        textures = glGenTextures(1)
        glBindTexture(GL_TEXTURE_EXTERNAL_OES, textures)
        glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
        glEGLImageTargetTexture2DOES(GL_TEXTURE_EXTERNAL_OES, image)

        return textures

    def resizeEvent(self, event):
        size = event.size()

        print('Resize', size)

        super().resizeEvent(event)

        if self.surface is None:
            return

        glViewport(0, 0, size.width() // 2, size.height())

    def paintEvent(self, event):
        if self.surface is None:
            self.init_gl()

        for ctx_idx, queue in self.reqqueue.items():
            if len(queue) == 0:
                continue

            ctx = next(ctx for ctx in self.state.contexts if ctx.idx == ctx_idx)

            if self.current[ctx_idx]:
                old = self.current[ctx_idx]
                self.current[ctx_idx] = None
                self.state.request_processed(ctx, old)

            next_req = queue.pop(0)
            self.current[ctx_idx] = next_req

            stream, fb = next(iter(next_req.buffers.items()))

            self.textures[stream] = self.create_texture(stream, fb)

        self.paint_gl()

    def paint_gl(self):
        b = eglMakeCurrent(self.egl.display, self.surface, self.surface, self.egl.context)
        assert(b)

        glClear(GL_COLOR_BUFFER_BIT)

        size = self.size()

        for idx, ctx in enumerate(self.state.contexts):
            for stream in ctx.streams:
                if self.textures[stream] is None:
                    continue

                w = size.width() // self.num_columns
                h = size.height() // self.num_rows

                x = idx % self.num_columns
                y = idx // self.num_columns

                x *= w
                y *= h

                glViewport(x, y, w, h)

                glBindTexture(GL_TEXTURE_EXTERNAL_OES, self.textures[stream])
                glDrawArrays(GL_TRIANGLE_FAN, 0, 4)

        b = eglSwapBuffers(self.egl.display, self.surface)
        assert(b)

    def handle_request(self, ctx, req):
        self.reqqueue[ctx.idx].append(req)
        self.update()



#!/usr/bin/env python3
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2022, Raspberry Pi Ltd
#
# alsc tuning tool

import sys

from ctt import *
from ctt_tools import parse_input

if __name__ == '__main__':
    """
    initialise calibration
    """
    if len(sys.argv) == 1:
        print("""
    PiSP Lens Shading Camera Tuning Tool version 1.0

    Required Arguments:
    '-i' : Calibration image directory.
    '-o' : Name of output json file.

    Optional Arguments:
    '-t' : Target platform - 'pisp' or 'vc4'. Default 'vc4'
    '-c' : Config file for the CTT. If not passed, default parameters used.
    '-l' : Name of output log file. If not passed, 'ctt_log.txt' used.
              """)
        quit(0)
    else:
        """
        parse input arguments
        """
        json_output, directory, config, log_output, target = parse_input()
        if target == 'pisp':
            from ctt_pisp import json_template, grid_size
        elif target == 'vc4':
            from ctt_vc4 import json_template, grid_size

        run_ctt(json_output, directory, config, log_output, json_template, grid_size, target, alsc_only=True)



# Copyright 2014 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import unittest

from mojom.generate import module as mojom
from mojom.generate import translate
from mojom.parse import ast

class TranslateTest(unittest.TestCase):
  """Tests |parser.Parse()|."""

  def testSimpleArray(self):
    """Tests a simple int32[]."""
    # pylint: disable=W0212
    self.assertEquals(translate._MapKind("int32[]"), "a:i32")

  def testAssociativeArray(self):
    """Tests a simple uint8{string}."""
    # pylint: disable=W0212
    self.assertEquals(translate._MapKind("uint8{string}"), "m[s][u8]")

  def testLeftToRightAssociativeArray(self):
    """Makes sure that parsing is done from right to left on the internal kinds
       in the presence of an associative array."""
    # pylint: disable=W0212
    self.assertEquals(translate._MapKind("uint8[]{string}"), "m[s][a:u8]")

  def testTranslateSimpleUnions(self):
    """Makes sure that a simple union is translated correctly."""
    tree = ast.Mojom(None, ast.ImportList(), [
        ast.Union(
            "SomeUnion", None,
            ast.UnionBody([
                ast.UnionField("a", None, None, "int32"),
                ast.UnionField("b", None, None, "string")
            ]))
    ])

    translation = translate.OrderedModule(tree, "mojom_tree", [])
    self.assertEqual(1, len(translation.unions))

    union = translation.unions[0]
    self.assertTrue(isinstance(union, mojom.Union))
    self.assertEqual("SomeUnion", union.mojom_name)
    self.assertEqual(2, len(union.fields))
    self.assertEqual("a", union.fields[0].mojom_name)
    self.assertEqual(mojom.INT32.spec, union.fields[0].kind.spec)
    self.assertEqual("b", union.fields[1].mojom_name)
    self.assertEqual(mojom.STRING.spec, union.fields[1].kind.spec)

  def testMapKindRaisesWithDuplicate(self):
    """Verifies _MapTreeForType() raises when passed two values with the same
       name."""
    methods = [
        ast.Method('dup', None, None, ast.ParameterList(), None),
        ast.Method('dup', None, None, ast.ParameterList(), None)
    ]
    with self.assertRaises(Exception):
      translate._ElemsOfType(methods, ast.Method, 'scope')

  def testAssociatedKinds(self):
    """Tests type spec translation of associated interfaces and requests."""
    # pylint: disable=W0212
    self.assertEquals(
        translate._MapKind("asso<SomeInterface>?"), "?asso:x:SomeInterface")
    self.assertEquals(translate._MapKind("rca<SomeInterface>?"),
                      "?rca:x:SomeInterface")

  def testSelfRecursiveUnions(self):
    """Verifies _UnionField() raises when a union is self-recursive."""
    tree = ast.Mojom(None, ast.ImportList(), [
        ast.Union("SomeUnion", None,
                  ast.UnionBody([ast.UnionField("a", None, None, "SomeUnion")]))
    ])
    with self.assertRaises(Exception):
      translate.OrderedModule(tree, "mojom_tree", [])

    tree = ast.Mojom(None, ast.ImportList(), [
        ast.Union(
            "SomeUnion", None,
            ast.UnionBody([ast.UnionField("a", None, None, "SomeUnion?")]))
    ])
    with self.assertRaises(Exception):
      translate.OrderedModule(tree, "mojom_tree", [])

  def testDuplicateAttributesException(self):
    tree = ast.Mojom(None, ast.ImportList(), [
        ast.Union(
            "FakeUnion",
            ast.AttributeList([
                ast.Attribute("key1", "value"),
                ast.Attribute("key1", "value")
            ]),
            ast.UnionBody([
                ast.UnionField("a", None, None, "int32"),
                ast.UnionField("b", None, None, "string")
            ]))
    ])
    with self.assertRaises(Exception):
      translate.OrderedModule(tree, "mojom_tree", [])

  def testEnumWithReservedValues(self):
    """Verifies that assigning reserved values to enumerators fails."""
    # -128 is reserved for the empty representation in WTF::HashTraits.
    tree = ast.Mojom(None, ast.ImportList(), [
        ast.Enum(
            "MyEnum", None,
            ast.EnumValueList([
                ast.EnumValue('kReserved', None, '-128'),
            ]))
    ])
    with self.assertRaises(Exception) as context:
      translate.OrderedModule(tree, "mojom_tree", [])
    self.assertIn("reserved for WTF::HashTrait", str(context.exception))

    # -127 is reserved for the deleted representation in WTF::HashTraits.
    tree = ast.Mojom(None, ast.ImportList(), [
        ast.Enum(
            "MyEnum", None,
            ast.EnumValueList([
                ast.EnumValue('kReserved', None, '-127'),
            ]))
    ])
    with self.assertRaises(Exception) as context:
      translate.OrderedModule(tree, "mojom_tree", [])
    self.assertIn("reserved for WTF::HashTrait", str(context.exception))

    # Implicitly assigning a reserved value should also fail.
    tree = ast.Mojom(None, ast.ImportList(), [
        ast.Enum(
            "MyEnum", None,
            ast.EnumValueList([
                ast.EnumValue('kNotReserved', None, '-129'),
                ast.EnumValue('kImplicitlyReserved', None, None),
            ]))
    ])
    with self.assertRaises(Exception) as context:
      translate.OrderedModule(tree, "mojom_tree", [])
    self.assertIn("reserved for WTF::HashTrait", str(context.exception))



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>
#
# Tuning script for raspberrypi, ALSC only

import sys

import libtuning as lt
from libtuning.parsers import RaspberryPiParser
from libtuning.generators import RaspberryPiOutput

from raspberrypi.alsc import ALSC

tuner = lt.Tuner('Raspberry Pi (ALSC only)')
tuner.add(ALSC)
tuner.set_input_parser(RaspberryPiParser())
tuner.set_output_formatter(RaspberryPiOutput())
tuner.set_output_order([ALSC])

if __name__ == '__main__':
    sys.exit(tuner.run(sys.argv))



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2024, Ideas On Board

import logging

from ..module import Module

from libtuning.ctt_awb import awb
import numpy as np

logger = logging.getLogger(__name__)


class AWB(Module):
    type = 'awb'
    hr_name = 'AWB (Base)'
    out_name = 'GenericAWB'

    def __init__(self, *, debug: list):
        super().__init__()

        self.debug = debug

    def do_calculation(self, images):
        logger.info('Starting AWB calculation')

        imgs = [img for img in images if img.macbeth is not None]

        gains, _, _ = awb(imgs, None, None, False)
        gains = np.reshape(gains, (-1, 3))

        return [{
                    'ct': int(v[0]),
                    'gains': [float(1.0 / v[1]), float(1.0 / v[2])]
                } for v in gains]



#!/usr/bin/env python3
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2023, Raspberry Pi (Trading) Ltd.
#
# cac_only.py - cac tuning tool


# This file allows you to tune only the chromatic aberration correction
# Specify any number of files in the command line args, and it shall iterate through
# and generate an averaged cac table from all the input images, which you can then
# input into your tuning file.

# Takes .dng files produced by the camera modules of the dots grid and calculates the chromatic abberation of each dot.
# Then takes each dot, and works out where it was in the image, and uses that to output a tables of the shifts
# across the whole image.

from PIL import Image
import numpy as np
import rawpy
import sys
import getopt

from ctt_cac import *


def cac(filelist, output_filepath, plot_results=False):
    np.set_printoptions(precision=3)
    np.set_printoptions(suppress=True)

    # Create arrays to hold all the dots data and their colour offsets
    red_shift = []  # Format is: [[Dot Center X, Dot Center Y, x shift, y shift]]
    blue_shift = []
    # Iterate through the files
    # Multiple files is reccomended to average out the lens aberration through rotations
    for file in filelist:
        print("\n Processing file " + str(file))
        # Read the raw RGB values from the .dng file
        with rawpy.imread(file) as raw:
            rgb = raw.postprocess()
            sizes = (raw.sizes)

        image_size = [sizes[2], sizes[3]]  # Image size, X, Y
        # Create a colour copy of the RGB values to use later in the calibration
        imout = Image.new(mode="RGB", size=image_size)
        rgb_image = np.array(imout)
        # The rgb values need reshaping from a 1d array to a 3d array to be worked with easily
        rgb.reshape((image_size[0], image_size[1], 3))
        rgb_image = rgb

        # Pass the RGB image through to the dots locating program
        # Returns an array of the dots (colour rectangles around the dots), and an array of their locations
        print("Finding dots")
        dots, dots_locations = find_dots_locations(rgb_image)

        # Now, analyse each dot. Work out the centroid of each colour channel, and use that to work out
        # by how far the chromatic aberration has shifted each channel
        print('Dots found: ' + str(len(dots)))

        for dot, dot_location in zip(dots, dots_locations):
            if len(dot) > 0:
                if (dot_location[0] > 0) and (dot_location[1] > 0):
                    ret = analyse_dot(dot, dot_location)
                    red_shift.append(ret[0])
                    blue_shift.append(ret[1])

    # Take our arrays of red shifts and locations, push them through to be interpolated into a 9x9 matrix
    # for the CAC block to handle and then store these as a .json file to be added to the camera
    # tuning file
    print("\nCreating output grid")
    rx, ry, bx, by = shifts_to_yaml(red_shift, blue_shift, image_size)

    print("CAC correction complete!")

    # The json format that we then paste into the tuning file (manually)
    sample = '''
    {
        "rpi.cac" :
        {
            "strength": 1.0,
            "lut_rx" : [
                   rx_vals
            ],
            "lut_ry" : [
                   ry_vals
            ],
            "lut_bx" : [
                   bx_vals
            ],
            "lut_by" : [
                   by_vals
            ]
        }
    }
    '''

    # Below, may look incorrect, however, the PiSP (standard) dimensions are flipped in comparison to
    # PIL image coordinate directions, hence why xr -> yr. Also, the shifts calculated are colour shifts,
    # and the PiSP block asks for the values it should shift (hence the * -1, to convert from colour shift to a pixel shift)
    sample = sample.replace("rx_vals", pprint_array(ry * -1))
    sample = sample.replace("ry_vals", pprint_array(rx * -1))
    sample = sample.replace("bx_vals", pprint_array(by * -1))
    sample = sample.replace("by_vals", pprint_array(bx * -1))
    print("Successfully converted to JSON")
    f = open(str(output_filepath), "w+")
    f.write(sample)
    f.close()
    print("Successfully written to json file")
    '''
    If you wish to see a plot of the colour channel shifts, add the -p or --plots option
    Can be a quick way of validating if the data/dots you've got are good, or if you need to
    change some parameters/take some better images
    '''
    if plot_results:
        plot_shifts(red_shift, blue_shift)


if __name__ == "__main__":
    argv = sys.argv
    # Detect the input and output file paths
    arg_output = "output.json"
    arg_help = "{0} -i <input> -o <output> -p <plot results>".format(argv[0])
    opts, args = getopt.getopt(argv[1:], "hi:o:p", ["help", "input=", "output=", "plot"])

    output_location = 0
    input_location = 0
    filelist = []
    plot_results = False
    for i in range(len(argv)):
        if ("-h") in argv[i]:
            print(arg_help)  # print the help message
            sys.exit(2)
        if "-o" in argv[i]:
            output_location = i
        if ".dng" in argv[i]:
            filelist.append(argv[i])
        if "-p" in argv[i]:
            plot_results = True

    arg_output = argv[output_location + 1]
    cac(filelist, arg_output, plot_results)



#!/usr/bin/env python3
# Copyright 2023 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
#
# This utility minifies JS files with terser.
#
# Instance of 'node' has no 'RunNode' member (no-member)
# pylint: disable=no-member

import argparse
import os
import sys

_HERE_PATH = os.path.dirname(__file__)
_SRC_PATH = os.path.normpath(os.path.join(_HERE_PATH, '..', '..', '..', '..'))
_CWD = os.getcwd()
sys.path.append(os.path.join(_SRC_PATH, 'third_party', 'node'))
import node
import node_modules


def MinifyFile(input_file, output_file):
  node.RunNode([
      node_modules.PathToTerser(), input_file, '--mangle', '--compress',
      '--comments', 'false', '--output', output_file
  ])


def main(argv):
  parser = argparse.ArgumentParser()
  parser.add_argument('--input', required=True)
  parser.add_argument('--output', required=True)
  args = parser.parse_args(argv)

  # Delete the output file if it already exists. It may be a sym link to the
  # input, because in non-optimized/pre-Terser builds the input file is copied
  # to the output location with gn copy().
  out_path = os.path.join(_CWD, args.output)
  if (os.path.exists(out_path)):
    os.remove(out_path)

  MinifyFile(os.path.join(_CWD, args.input), out_path)


if __name__ == '__main__':
  main(sys.argv[1:])



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2024, Ideas on Board
#
# Module implementation for static data

from .module import Module


# This module can be used in cases where the tuning file should contain
# static data.
class StaticModule(Module):
    def __init__(self, out_name: str, output: dict = {}):
        super().__init__()
        self.out_name = out_name
        self.hr_name = f'Static {out_name}'
        self.type = f'static_{out_name}'
        self.output = output

    def validate_config(self, config: dict) -> bool:
        return True

    def process(self, config: dict, images: list, outputs: dict) -> dict:
        return self.output



# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2022, Tomi Valkeinen <tomi.valkeinen@ideasonboard.com>

from OpenGL.EGL.VERSION.EGL_1_0 import EGLNativeDisplayType, eglGetProcAddress, eglQueryString, EGL_EXTENSIONS

from OpenGL.raw.GLES2 import _types as _cs
from OpenGL.GLES2.VERSION.GLES2_2_0 import *
from OpenGL.GLES3.VERSION.GLES3_3_0 import *
from OpenGL import GL as gl

from ctypes import c_int, c_char_p, c_void_p, cdll, POINTER, util, \
    pointer, CFUNCTYPE, c_bool


def getEGLNativeDisplay():
    _x11lib = cdll.LoadLibrary(util.find_library('X11'))
    XOpenDisplay = _x11lib.XOpenDisplay
    XOpenDisplay.argtypes = [c_char_p]
    XOpenDisplay.restype = POINTER(EGLNativeDisplayType)

    return XOpenDisplay(None)


# Hack. PyOpenGL doesn't seem to manage to find glEGLImageTargetTexture2DOES.
def getglEGLImageTargetTexture2DOES():
    funcptr = eglGetProcAddress('glEGLImageTargetTexture2DOES')
    prototype = CFUNCTYPE(None, _cs.GLenum, _cs.GLeglImageOES)
    return prototype(funcptr)


glEGLImageTargetTexture2DOES = getglEGLImageTargetTexture2DOES()


def get_gl_extensions():
    n = GLint()
    glGetIntegerv(GL_NUM_EXTENSIONS, n)
    gl_extensions = []
    for i in range(n.value):
        gl_extensions.append(gl.glGetStringi(GL_EXTENSIONS, i).decode())
    return gl_extensions


def check_gl_extensions(required_extensions):
    extensions = get_gl_extensions()

    if False:
        print('GL EXTENSIONS: ', ' '.join(extensions))

    for ext in required_extensions:
        if ext not in extensions:
            raise Exception(ext + ' missing')


def get_egl_extensions(egl_display):
    return eglQueryString(egl_display, EGL_EXTENSIONS).decode().split(' ')


def check_egl_extensions(egl_display, required_extensions):
    extensions = get_egl_extensions(egl_display)

    if False:
        print('EGL EXTENSIONS: ', ' '.join(extensions))

    for ext in required_extensions:
        if ext not in extensions:
            raise Exception(ext + ' missing')



# Copyright 2020 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from mojom.generate import module
from mojom_parser_test_case import MojomParserTestCase


class VersionCompatibilityTest(MojomParserTestCase):
  """Tests covering compatibility between two versions of the same mojom type
  definition. This coverage ensures that we can reliably detect unsafe changes
  to definitions that are expected to tolerate version skew in production
  environments."""

  def _GetTypeCompatibilityMap(self, old_mojom, new_mojom):
    """Helper to support the implementation of assertBackwardCompatible and
    assertNotBackwardCompatible."""

    old = self.ExtractTypes(old_mojom)
    new = self.ExtractTypes(new_mojom)
    self.assertEqual(set(old.keys()), set(new.keys()),
                     'Old and new test mojoms should use the same type names.')

    checker = module.BackwardCompatibilityChecker()
    compatibility_map = {}
    for name in old:
      try:
        compatibility_map[name] = checker.IsBackwardCompatible(
            new[name], old[name])
      except Exception:
        compatibility_map[name] = False
    return compatibility_map

  def assertBackwardCompatible(self, old_mojom, new_mojom):
    compatibility_map = self._GetTypeCompatibilityMap(old_mojom, new_mojom)
    for name, compatible in compatibility_map.items():
      if not compatible:
        raise AssertionError(
            'Given the old mojom:\n\n    %s\n\nand the new mojom:\n\n    %s\n\n'
            'The new definition of %s should pass a backward-compatibiity '
            'check, but it does not.' % (old_mojom, new_mojom, name))

  def assertNotBackwardCompatible(self, old_mojom, new_mojom):
    compatibility_map = self._GetTypeCompatibilityMap(old_mojom, new_mojom)
    if all(compatibility_map.values()):
      raise AssertionError(
          'Given the old mojom:\n\n    %s\n\nand the new mojom:\n\n    %s\n\n'
          'The new mojom should fail a backward-compatibility check, but it '
          'does not.' % (old_mojom, new_mojom))

  def testNewNonExtensibleEnumValue(self):
    """Adding a value to a non-extensible enum breaks backward-compatibility."""
    self.assertNotBackwardCompatible('enum E { kFoo, kBar };',
                                     'enum E { kFoo, kBar, kBaz };')

  def testNewNonExtensibleEnumValueWithMinVersion(self):
    """Adding a value to a non-extensible enum breaks backward-compatibility,
    even with a new [MinVersion] specified for the value."""
    self.assertNotBackwardCompatible(
        'enum E { kFoo, kBar };', 'enum E { kFoo, kBar, [MinVersion=1] kBaz };')

  def testNewValueInExistingVersion(self):
    """Adding a value to an existing version is not allowed, even if the old
    enum was marked [Extensible]. Note that it is irrelevant whether or not the
    new enum is marked [Extensible]."""
    self.assertNotBackwardCompatible(
        '[Extensible] enum E { [Default] kFoo, kBar };',
        'enum E { kFoo, kBar, kBaz };')
    self.assertNotBackwardCompatible(
        '[Extensible] enum E { [Default] kFoo, kBar };',
        '[Extensible] enum E { [Default] kFoo, kBar, kBaz };')
    self.assertNotBackwardCompatible(
        '[Extensible] enum E { [Default] kFoo, [MinVersion=1] kBar };',
        'enum E { kFoo, [MinVersion=1] kBar, [MinVersion=1] kBaz };')

  def testEnumValueRemoval(self):
    """Removal of an enum value is never valid even for [Extensible] enums."""
    self.assertNotBackwardCompatible('enum E { kFoo, kBar };',
                                     'enum E { kFoo };')
    self.assertNotBackwardCompatible(
        '[Extensible] enum E { [Default] kFoo, kBar };',
        '[Extensible] enum E { [Default] kFoo };')
    self.assertNotBackwardCompatible(
        '[Extensible] enum E { [Default] kA, [MinVersion=1] kB };',
        '[Extensible] enum E { [Default] kA, };')
    self.assertNotBackwardCompatible(
        """[Extensible] enum E {
          [Default] kA,
          [MinVersion=1] kB,
          [MinVersion=1] kZ };""",
        '[Extensible] enum E { [Default] kA, [MinVersion=1] kB };')

  def testNewExtensibleEnumValueWithMinVersion(self):
    """Adding a new and properly [MinVersion]'d value to an [Extensible] enum
    is a backward-compatible change. Note that it is irrelevant whether or not
    the new enum is marked [Extensible]."""
    self.assertBackwardCompatible('[Extensible] enum E { [Default] kA, kB };',
                                  'enum E { kA, kB, [MinVersion=1] kC };')
    self.assertBackwardCompatible(
        '[Extensible] enum E { [Default] kA, kB };',
        '[Extensible] enum E { [Default] kA, kB, [MinVersion=1] kC };')
    self.assertBackwardCompatible(
        '[Extensible] enum E { [Default] kA, [MinVersion=1] kB };',
        """[Extensible] enum E {
          [Default] kA,
          [MinVersion=1] kB,
          [MinVersion=2] kC };""")

  def testRenameEnumValue(self):
    """Renaming an enum value does not affect backward-compatibility. Only
    numeric value is relevant."""
    self.assertBackwardCompatible('enum E { kA, kB };', 'enum E { kX, kY };')

  def testAddEnumValueAlias(self):
    """Adding new enum fields does not affect backward-compatibility if it does
    not introduce any new numeric values."""
    self.assertBackwardCompatible(
        'enum E { kA, kB };', 'enum E { kA, kB, kC = kA, kD = 1, kE = kD };')

  def testEnumIdentity(self):
    """An unchanged enum is obviously backward-compatible."""
    self.assertBackwardCompatible('enum E { kA, kB, kC };',
                                  'enum E { kA, kB, kC };')

  def testNewStructFieldUnversioned(self):
    """Adding a new field to a struct without a new (i.e. higher than any
    existing version) [MinVersion] tag breaks backward-compatibility."""
    self.assertNotBackwardCompatible('struct S { string a; };',
                                     'struct S { string a; string b; };')

  def testStructFieldRemoval(self):
    """Removing a field from a struct breaks backward-compatibility."""
    self.assertNotBackwardCompatible('struct S { string a; string b; };',
                                     'struct S { string a; };')

  def testStructFieldTypeChange(self):
    """Changing the type of an existing field always breaks
    backward-compatibility."""
    self.assertNotBackwardCompatible('struct S { string a; };',
                                     'struct S { array<int32> a; };')

  def testStructFieldBecomingOptional(self):
    """Changing a field from non-optional to optional breaks
    backward-compatibility."""
    self.assertNotBackwardCompatible('struct S { string a; };',
                                     'struct S { string? a; };')

  def testStructFieldBecomingNonOptional(self):
    """Changing a field from optional to non-optional breaks
    backward-compatibility."""
    self.assertNotBackwardCompatible('struct S { string? a; };',
                                     'struct S { string a; };')

  def testStructFieldOrderChange(self):
    """Changing the order of fields breaks backward-compatibility."""
    self.assertNotBackwardCompatible('struct S { string a; bool b; };',
                                     'struct S { bool b; string a; };')
    self.assertNotBackwardCompatible('struct S { string a@0; bool b@1; };',
                                     'struct S { string a@1; bool b@0; };')

  def testStructFieldMinVersionChange(self):
    """Changing the MinVersion of a field breaks backward-compatibility."""
    self.assertNotBackwardCompatible(
        'struct S { string a; [MinVersion=1] string? b; };',
        'struct S { string a; [MinVersion=2] string? b; };')

  def testStructFieldTypeChange(self):
    """If a struct field's own type definition changes, the containing struct
    is backward-compatible if and only if the field type's change is
    backward-compatible."""
    self.assertBackwardCompatible(
        'struct S {}; struct T { S s; };',
        'struct S { [MinVersion=1] int32 x; }; struct T { S s; };')
    self.assertBackwardCompatible(
        '[Extensible] enum E { [Default] kA }; struct S { E e; };',
        """[Extensible] enum E {
          [Default] kA,
          [MinVersion=1] kB };
          struct S { E e; };""")
    self.assertNotBackwardCompatible(
        'struct S {}; struct T { S s; };',
        'struct S { int32 x; }; struct T { S s; };')
    self.assertNotBackwardCompatible(
        '[Extensible] enum E { [Default] kA }; struct S { E e; };',
        '[Extensible] enum E { [Default] kA, kB }; struct S { E e; };')

  def testNewStructFieldWithInvalidMinVersion(self):
    """Adding a new field using an existing MinVersion breaks backward-
    compatibility."""
    self.assertNotBackwardCompatible(
        """\
        struct S {
          string a;
          [MinVersion=1] string? b;
        };
        """, """\
        struct S {
          string a;
          [MinVersion=1] string? b;
          [MinVersion=1] string? c;
        };""")

  def testNewStructFieldWithValidMinVersion(self):
    """Adding a new field is safe if tagged with a MinVersion greater than any
    previously used MinVersion in the struct."""
    self.assertBackwardCompatible(
        'struct S { int32 a; };',
        'struct S { int32 a; [MinVersion=1] int32 b; };')
    self.assertBackwardCompatible(
        'struct S { int32 a; [MinVersion=1] int32 b; };',
        'struct S { int32 a; [MinVersion=1] int32 b; [MinVersion=2] bool c; };')

  def testNewStructFieldNullableReference(self):
    """Adding a new nullable reference-typed field is fine if versioned
    properly."""
    self.assertBackwardCompatible(
        'struct S { int32 a; };',
        'struct S { int32 a; [MinVersion=1] string? b; };')

  def testStructFieldRename(self):
    """Renaming a field has no effect on backward-compatibility."""
    self.assertBackwardCompatible('struct S { int32 x; bool b; };',
                                  'struct S { int32 a; bool b; };')

  def testStructFieldReorderWithExplicitOrdinals(self):
    """Reordering fields has no effect on backward-compatibility when field
    ordinals are explicitly labeled and remain unchanged."""
    self.assertBackwardCompatible('struct S { bool b@1; int32 a@0; };',
                                  'struct S { int32 a@0; bool b@1; };')

  def testNewUnionFieldUnversioned(self):
    """Adding a new field to a union without a new (i.e. higher than any
    existing version) [MinVersion] tag breaks backward-compatibility."""
    self.assertNotBackwardCompatible('union U { string a; };',
                                     'union U { string a; string b; };')

  def testUnionFieldRemoval(self):
    """Removing a field from a union breaks backward-compatibility."""
    self.assertNotBackwardCompatible('union U { string a; string b; };',
                                     'union U { string a; };')

  def testUnionFieldTypeChange(self):
    """Changing the type of an existing field always breaks
    backward-compatibility."""
    self.assertNotBackwardCompatible('union U { string a; };',
                                     'union U { array<int32> a; };')

  def testUnionFieldBecomingOptional(self):
    """Changing a field from non-optional to optional breaks
    backward-compatibility."""
    self.assertNotBackwardCompatible('union U { string a; };',
                                     'union U { string? a; };')

  def testFieldNestedTypeChanged(self):
    """Changing the definition of a nested type within a field (such as an array
    element or interface endpoint type) should only break backward-compatibility
    if the changes to that type are not backward-compatible."""
    self.assertBackwardCompatible(
        """\
        struct S { string a; };
        struct T { array<S> ss; };
        """, """\
        struct S {
          string a;
          [MinVersion=1] string? b;
        };
        struct T { array<S> ss; };
        """)
    self.assertBackwardCompatible(
        """\
        interface F { Do(); };
        struct S { pending_receiver<F> r; };
        """, """\
        interface F {
          Do();
          [MinVersion=1] Say();
        };
        struct S { pending_receiver<F> r; };
        """)

  def testRecursiveTypeChange(self):
    """Recursive types do not break the compatibility checker."""
    self.assertBackwardCompatible(
        """\
        struct S {
          string a;
          array<S> others;
        };""", """\
        struct S {
          string a;
          array<S> others;
          [MinVersion=1] string? b;
        };""")

  def testUnionFieldBecomingNonOptional(self):
    """Changing a field from optional to non-optional breaks
    backward-compatibility."""
    self.assertNotBackwardCompatible('union U { string? a; };',
                                     'union U { string a; };')

  def testUnionFieldOrderChange(self):
    """Changing the order of fields breaks backward-compatibility."""
    self.assertNotBackwardCompatible('union U { string a; bool b; };',
                                     'union U { bool b; string a; };')
    self.assertNotBackwardCompatible('union U { string a@0; bool b@1; };',
                                     'union U { string a@1; bool b@0; };')

  def testUnionFieldMinVersionChange(self):
    """Changing the MinVersion of a field breaks backward-compatibility."""
    self.assertNotBackwardCompatible(
        'union U { string a; [MinVersion=1] string b; };',
        'union U { string a; [MinVersion=2] string b; };')

  def testUnionFieldTypeChange(self):
    """If a union field's own type definition changes, the containing union
    is backward-compatible if and only if the field type's change is
    backward-compatible."""
    self.assertBackwardCompatible(
        'struct S {}; union U { S s; };',
        'struct S { [MinVersion=1] int32 x; }; union U { S s; };')
    self.assertBackwardCompatible(
        '[Extensible] enum E { [Default] kA }; union U { E e; };',
        """[Extensible] enum E {
          [Default] kA,
          [MinVersion=1] kB };
          union U { E e; };""")
    self.assertNotBackwardCompatible(
        'struct S {}; union U { S s; };',
        'struct S { int32 x; }; union U { S s; };')
    self.assertNotBackwardCompatible(
        '[Extensible] enum E { [Default] kA }; union U { E e; };',
        '[Extensible] enum E { [Default] kA, kB }; union U { E e; };')

  def testNewUnionFieldWithInvalidMinVersion(self):
    """Adding a new field using an existing MinVersion breaks backward-
    compatibility."""
    self.assertNotBackwardCompatible(
        """\
        union U {
          string a;
          [MinVersion=1] string b;
        };
        """, """\
        union U {
          string a;
          [MinVersion=1] string b;
          [MinVersion=1] string c;
        };""")

  def testNewUnionFieldWithValidMinVersion(self):
    """Adding a new field is safe if tagged with a MinVersion greater than any
    previously used MinVersion in the union."""
    self.assertBackwardCompatible(
        'union U { int32 a; };',
        'union U { int32 a; [MinVersion=1] int32 b; };')
    self.assertBackwardCompatible(
        'union U { int32 a; [MinVersion=1] int32 b; };',
        'union U { int32 a; [MinVersion=1] int32 b; [MinVersion=2] bool c; };')

  def testUnionFieldRename(self):
    """Renaming a field has no effect on backward-compatibility."""
    self.assertBackwardCompatible('union U { int32 x; bool b; };',
                                  'union U { int32 a; bool b; };')

  def testUnionFieldReorderWithExplicitOrdinals(self):
    """Reordering fields has no effect on backward-compatibility when field
    ordinals are explicitly labeled and remain unchanged."""
    self.assertBackwardCompatible('union U { bool b@1; int32 a@0; };',
                                  'union U { int32 a@0; bool b@1; };')

  def testNewInterfaceMethodUnversioned(self):
    """Adding a new method to an interface without a new (i.e. higher than any
    existing version) [MinVersion] tag breaks backward-compatibility."""
    self.assertNotBackwardCompatible('interface F { A(); };',
                                     'interface F { A(); B(); };')

  def testInterfaceMethodRemoval(self):
    """Removing a method from an interface breaks backward-compatibility."""
    self.assertNotBackwardCompatible('interface F { A(); B(); };',
                                     'interface F { A(); };')

  def testInterfaceMethodParamsChanged(self):
    """Changes to the parameter list are only backward-compatible if they meet
    backward-compatibility requirements of an equivalent struct definition."""
    self.assertNotBackwardCompatible('interface F { A(); };',
                                     'interface F { A(int32 x); };')
    self.assertNotBackwardCompatible('interface F { A(int32 x); };',
                                     'interface F { A(bool x); };')
    self.assertNotBackwardCompatible(
        'interface F { A(int32 x, [MinVersion=1] string? s); };', """\
        interface F {
          A(int32 x, [MinVersion=1] string? s, [MinVersion=1] int32 y);
        };""")

    self.assertBackwardCompatible('interface F { A(int32 x); };',
                                  'interface F { A(int32 a); };')
    self.assertBackwardCompatible(
        'interface F { A(int32 x); };',
        'interface F { A(int32 x, [MinVersion=1] string? s); };')

    self.assertBackwardCompatible(
        'struct S {}; interface F { A(S s); };',
        'struct S { [MinVersion=1] int32 x; }; interface F { A(S s); };')
    self.assertBackwardCompatible(
        'struct S {}; struct T {}; interface F { A(S s); };',
        'struct S {}; struct T {}; interface F { A(T s); };')
    self.assertNotBackwardCompatible(
        'struct S {}; struct T { int32 x; }; interface F { A(S s); };',
        'struct S {}; struct T { int32 x; }; interface F { A(T t); };')

  def testInterfaceMethodReplyAdded(self):
    """Adding a reply to a message breaks backward-compatibilty."""
    self.assertNotBackwardCompatible('interface F { A(); };',
                                     'interface F { A() => (); };')

  def testInterfaceMethodReplyRemoved(self):
    """Removing a reply from a message breaks backward-compatibility."""
    self.assertNotBackwardCompatible('interface F { A() => (); };',
                                     'interface F { A(); };')

  def testInterfaceMethodReplyParamsChanged(self):
    """Similar to request parameters, a change to reply parameters is considered
    backward-compatible if it meets the same backward-compatibility
    requirements imposed on equivalent struct changes."""
    self.assertNotBackwardCompatible('interface F { A() => (); };',
                                     'interface F { A() => (int32 x); };')
    self.assertNotBackwardCompatible('interface F { A() => (int32 x); };',
                                     'interface F { A() => (); };')
    self.assertNotBackwardCompatible('interface F { A() => (bool x); };',
                                     'interface F { A() => (int32 x); };')

    self.assertBackwardCompatible('interface F { A() => (int32 a); };',
                                  'interface F { A() => (int32 x); };')
    self.assertBackwardCompatible(
        'interface F { A() => (int32 x); };',
        'interface F { A() => (int32 x, [MinVersion] string? s); };')

  def testNewInterfaceMethodWithInvalidMinVersion(self):
    """Adding a new method to an existing version is not backward-compatible."""
    self.assertNotBackwardCompatible(
        """\
        interface F {
          A();
          [MinVersion=1] B();
        };
        """, """\
        interface F {
          A();
          [MinVersion=1] B();
          [MinVersion=1] C();
        };
        """)

  def testNewInterfaceMethodWithValidMinVersion(self):
    """Adding a new method is fine as long as its MinVersion exceeds that of any
    method on the old interface definition."""
    self.assertBackwardCompatible('interface F { A(); };',
                                  'interface F { A(); [MinVersion=1] B(); };')



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
#
# camera tuning tool for lux level

from ctt_tools import *


"""
Find lux values from metadata and calculate Y
"""
def lux(Cam, Img):
    shutter_speed = Img.exposure
    gain = Img.againQ8_norm
    aperture = 1
    Cam.log += '\nShutter speed = {}'.format(shutter_speed)
    Cam.log += '\nGain = {}'.format(gain)
    Cam.log += '\nAperture = {}'.format(aperture)
    patches = [Img.patches[i] for i in Img.order]
    channels = [Img.channels[i] for i in Img.order]
    return lux_calc(Cam, Img, patches, channels), shutter_speed, gain


"""
perform lux calibration on bayer channels
"""
def lux_calc(Cam, Img, patches, channels):
    """
    find means color channels on grey patches
    """
    ap_r = np.mean(patches[0][3::4])
    ap_g = (np.mean(patches[1][3::4])+np.mean(patches[2][3::4]))/2
    ap_b = np.mean(patches[3][3::4])
    Cam.log += '\nAverage channel values on grey patches:'
    Cam.log += '\nRed = {:.0f} Green = {:.0f} Blue = {:.0f}'.format(ap_r, ap_b, ap_g)
    # print(ap_r, ap_g, ap_b)
    """
    calculate channel gains
    """
    gr = ap_g/ap_r
    gb = ap_g/ap_b
    Cam.log += '\nChannel gains: Red = {:.3f} Blue = {:.3f}'.format(gr, gb)

    """
    find means color channels on image and scale by gain
    note greens are averaged together (treated as one channel)
    """
    a_r = np.mean(channels[0])*gr
    a_g = (np.mean(channels[1])+np.mean(channels[2]))/2
    a_b = np.mean(channels[3])*gb
    Cam.log += '\nAverage channel values over entire image scaled by channel gains:'
    Cam.log += '\nRed = {:.0f} Green = {:.0f} Blue = {:.0f}'.format(a_r, a_b, a_g)
    # print(a_r, a_g, a_b)
    """
    Calculate y with top row of yuv matrix
    """
    y = 0.299*a_r + 0.587*a_g + 0.114*a_b
    Cam.log += '\nY value calculated: {}'.format(int(y))
    # print(y)
    return int(y)



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2023, Raspberry Pi Ltd
#
# ctt_cac.py - CAC (Chromatic Aberration Correction) tuning tool

from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

from ctt_dots_locator import find_dots_locations


# This is the wrapper file that creates a JSON entry for you to append
# to your camera tuning file.
# It calculates the chromatic aberration at different points throughout
# the image and uses that to produce a martix that can then be used
# in the camera tuning files to correct this aberration.


def pprint_array(array):
    # Function to print the array in a tidier format
    array = array
    output = ""
    for i in range(len(array)):
        for j in range(len(array[0])):
            output += str(round(array[i, j], 2)) + ", "
        # Add the necessary indentation to the array
        output += "\n                   "
    # Cut off the end of the array (nicely formats it)
    return output[:-22]


def plot_shifts(red_shifts, blue_shifts):
    # If users want, they can pass a command line option to show the shifts on a graph
    # Can be useful to check that the functions are all working, and that the sample
    # images are doing the right thing
    Xs = np.array(red_shifts)[:, 0]
    Ys = np.array(red_shifts)[:, 1]
    Zs = np.array(red_shifts)[:, 2]
    Zs2 = np.array(red_shifts)[:, 3]
    Zs3 = np.array(blue_shifts)[:, 2]
    Zs4 = np.array(blue_shifts)[:, 3]

    fig, axs = plt.subplots(2, 2)
    ax = fig.add_subplot(2, 2, 1, projection='3d')
    ax.scatter(Xs, Ys, Zs, cmap=cm.jet, linewidth=0)
    ax.set_title('Red X Shift')
    ax = fig.add_subplot(2, 2, 2, projection='3d')
    ax.scatter(Xs, Ys, Zs2, cmap=cm.jet, linewidth=0)
    ax.set_title('Red Y Shift')
    ax = fig.add_subplot(2, 2, 3, projection='3d')
    ax.scatter(Xs, Ys, Zs3, cmap=cm.jet, linewidth=0)
    ax.set_title('Blue X Shift')
    ax = fig.add_subplot(2, 2, 4, projection='3d')
    ax.scatter(Xs, Ys, Zs4, cmap=cm.jet, linewidth=0)
    ax.set_title('Blue Y Shift')
    fig.tight_layout()
    plt.show()


def shifts_to_yaml(red_shift, blue_shift, image_dimensions, output_grid_size=9):
    # Convert the shifts to a numpy array for easier handling and initialise other variables
    red_shifts = np.array(red_shift)
    blue_shifts = np.array(blue_shift)
    # create a grid that's smaller than the output grid, which we then interpolate from to get the output values
    xrgrid = np.zeros((output_grid_size - 1, output_grid_size - 1))
    xbgrid = np.zeros((output_grid_size - 1, output_grid_size - 1))
    yrgrid = np.zeros((output_grid_size - 1, output_grid_size - 1))
    ybgrid = np.zeros((output_grid_size - 1, output_grid_size - 1))

    xrsgrid = []
    xbsgrid = []
    yrsgrid = []
    ybsgrid = []
    xg = np.zeros((output_grid_size - 1, output_grid_size - 1))
    yg = np.zeros((output_grid_size - 1, output_grid_size - 1))

    # Format the grids - numpy doesn't work for this, it wants a
    # nice uniformly spaced grid, which we don't know if we have yet, hence the rather mundane setup
    for x in range(output_grid_size - 1):
        xrsgrid.append([])
        yrsgrid.append([])
        xbsgrid.append([])
        ybsgrid.append([])
        for y in range(output_grid_size - 1):
            xrsgrid[x].append([])
            yrsgrid[x].append([])
            xbsgrid[x].append([])
            ybsgrid[x].append([])

    image_size = (image_dimensions[0], image_dimensions[1])
    gridxsize = image_size[0] / (output_grid_size - 1)
    gridysize = image_size[1] / (output_grid_size - 1)

    # Iterate through each dot, and it's shift values and put these into the correct grid location
    for red_shift in red_shifts:
        xgridloc = int(red_shift[0] / gridxsize)
        ygridloc = int(red_shift[1] / gridysize)
        xrsgrid[xgridloc][ygridloc].append(red_shift[2])
        yrsgrid[xgridloc][ygridloc].append(red_shift[3])

    for blue_shift in blue_shifts:
        xgridloc = int(blue_shift[0] / gridxsize)
        ygridloc = int(blue_shift[1] / gridysize)
        xbsgrid[xgridloc][ygridloc].append(blue_shift[2])
        ybsgrid[xgridloc][ygridloc].append(blue_shift[3])

    # Now calculate the average pixel shift for each square in the grid
    for x in range(output_grid_size - 1):
        for y in range(output_grid_size - 1):
            xrgrid[x, y] = np.mean(xrsgrid[x][y])
            yrgrid[x, y] = np.mean(yrsgrid[x][y])
            xbgrid[x, y] = np.mean(xbsgrid[x][y])
            ybgrid[x, y] = np.mean(ybsgrid[x][y])

    # Next, we start to interpolate the central points of the grid that gets passed to the tuning file
    input_grids = np.array([xrgrid, yrgrid, xbgrid, ybgrid])
    output_grids = np.zeros((4, output_grid_size, output_grid_size))

    # Interpolate the centre of the grid
    output_grids[:, 1:-1, 1:-1] = (input_grids[:, 1:, :-1] + input_grids[:, 1:, 1:] + input_grids[:, :-1, 1:] + input_grids[:, :-1, :-1]) / 4

    # Edge cases:
    output_grids[:, 1:-1, 0] = ((input_grids[:, :-1, 0] + input_grids[:, 1:, 0]) / 2 - output_grids[:, 1:-1, 1]) * 2 + output_grids[:, 1:-1, 1]
    output_grids[:, 1:-1, -1] = ((input_grids[:, :-1, 7] + input_grids[:, 1:, 7]) / 2 - output_grids[:, 1:-1, -2]) * 2 + output_grids[:, 1:-1, -2]
    output_grids[:, 0, 1:-1] = ((input_grids[:, 0, :-1] + input_grids[:, 0, 1:]) / 2 - output_grids[:, 1, 1:-1]) * 2 + output_grids[:, 1, 1:-1]
    output_grids[:, -1, 1:-1] = ((input_grids[:, 7, :-1] + input_grids[:, 7, 1:]) / 2 - output_grids[:, -2, 1:-1]) * 2 + output_grids[:, -2, 1:-1]

    # Corner Cases:
    output_grids[:, 0, 0] = (output_grids[:, 0, 1] - output_grids[:, 1, 1]) + (output_grids[:, 1, 0] - output_grids[:, 1, 1]) + output_grids[:, 1, 1]
    output_grids[:, 0, -1] = (output_grids[:, 0, -2] - output_grids[:, 1, -2]) + (output_grids[:, 1, -1] - output_grids[:, 1, -2]) + output_grids[:, 1, -2]
    output_grids[:, -1, 0] = (output_grids[:, -1, 1] - output_grids[:, -2, 1]) + (output_grids[:, -2, 0] - output_grids[:, -2, 1]) + output_grids[:, -2, 1]
    output_grids[:, -1, -1] = (output_grids[:, -2, -1] - output_grids[:, -2, -2]) + (output_grids[:, -1, -2] - output_grids[:, -2, -2]) + output_grids[:, -2, -2]

    # Below, we swap the x and the y coordinates, and also multiply by a factor of -1
    # This is due to the PiSP (standard) dimensions being flipped in comparison to
    # PIL image coordinate directions, hence why xr -> yr. Also, the shifts calculated are colour shifts,
    # and the PiSP block asks for the values it should shift by (hence the * -1, to convert from colour shift to a pixel shift)

    output_grid_yr, output_grid_xr, output_grid_yb, output_grid_xb = output_grids * -1
    return output_grid_xr, output_grid_yr, output_grid_xb, output_grid_yb


def analyse_dot(dot, dot_location=[0, 0]):
    # Scan through the dot, calculate the centroid of each colour channel by doing:
    # pixel channel brightness * distance from top left corner
    # Sum these, and divide by the sum of each channel's brightnesses to get a centroid for each channel
    red_channel = np.array(dot)[:, :, 0]
    y_num_pixels = len(red_channel[0])
    x_num_pixels = len(red_channel)
    yred_weight = np.sum(np.dot(red_channel, np.arange(y_num_pixels)))
    xred_weight = np.sum(np.dot(np.arange(x_num_pixels), red_channel))
    red_sum = np.sum(red_channel)

    green_channel = np.array(dot)[:, :, 1]
    ygreen_weight = np.sum(np.dot(green_channel, np.arange(y_num_pixels)))
    xgreen_weight = np.sum(np.dot(np.arange(x_num_pixels), green_channel))
    green_sum = np.sum(green_channel)

    blue_channel = np.array(dot)[:, :, 2]
    yblue_weight = np.sum(np.dot(blue_channel, np.arange(y_num_pixels)))
    xblue_weight = np.sum(np.dot(np.arange(x_num_pixels), blue_channel))
    blue_sum = np.sum(blue_channel)

    # We return this structure. It contains 2 arrays that contain:
    # the locations of the dot center, along with the channel shifts in the x and y direction:
    # [ [red_center_x, red_center_y, red_x_shift, red_y_shift], [blue_center_x, blue_center_y, blue_x_shift, blue_y_shift] ]

    return [[int(dot_location[0]) + int(len(dot) / 2), int(dot_location[1]) + int(len(dot[0]) / 2), xred_weight / red_sum - xgreen_weight / green_sum, yred_weight / red_sum - ygreen_weight / green_sum], [dot_location[0] + int(len(dot) / 2), dot_location[1] + int(len(dot[0]) / 2), xblue_weight / blue_sum - xgreen_weight / green_sum, yblue_weight / blue_sum - ygreen_weight / green_sum]]


def cac(Cam):
    filelist = Cam.imgs_cac

    Cam.log += '\nCAC analysing files: {}'.format(str(filelist))
    np.set_printoptions(precision=3)
    np.set_printoptions(suppress=True)

    # Create arrays to hold all the dots data and their colour offsets
    red_shift = []  # Format is: [[Dot Center X, Dot Center Y, x shift, y shift]]
    blue_shift = []
    # Iterate through the files
    # Multiple files is reccomended to average out the lens aberration through rotations
    for file in filelist:
        Cam.log += '\nCAC processing file'
        print("\n Processing file")
        # Read the raw RGB values
        rgb = file.rgb
        image_size = [file.h, file.w]  # Image size, X, Y
        # Create a colour copy of the RGB values to use later in the calibration
        imout = Image.new(mode="RGB", size=image_size)
        rgb_image = np.array(imout)
        # The rgb values need reshaping from a 1d array to a 3d array to be worked with easily
        rgb.reshape((image_size[0], image_size[1], 3))
        rgb_image = rgb

        # Pass the RGB image through to the dots locating program
        # Returns an array of the dots (colour rectangles around the dots), and an array of their locations
        print("Finding dots")
        Cam.log += '\nFinding dots'
        dots, dots_locations = find_dots_locations(rgb_image)

        # Now, analyse each dot. Work out the centroid of each colour channel, and use that to work out
        # by how far the chromatic aberration has shifted each channel
        Cam.log += '\nDots found: {}'.format(str(len(dots)))
        print('Dots found: ' + str(len(dots)))

        for dot, dot_location in zip(dots, dots_locations):
            if len(dot) > 0:
                if (dot_location[0] > 0) and (dot_location[1] > 0):
                    ret = analyse_dot(dot, dot_location)
                    red_shift.append(ret[0])
                    blue_shift.append(ret[1])

    # Take our arrays of red shifts and locations, push them through to be interpolated into a 9x9 matrix
    # for the CAC block to handle and then store these as a .json file to be added to the camera
    # tuning file
    print("\nCreating output grid")
    Cam.log += '\nCreating output grid'
    rx, ry, bx, by = shifts_to_yaml(red_shift, blue_shift, image_size)

    print("CAC correction complete!")
    Cam.log += '\nCAC correction complete!'

    # Give the JSON dict back to the main ctt program
    return {"strength": 1.0, "lut_rx": list(rx.round(2).reshape(81)), "lut_ry": list(ry.round(2).reshape(81)), "lut_bx": list(bx.round(2).reshape(81)), "lut_by": list(by.round(2).reshape(81))}



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2019, Google Inc.
# Copyright (C) 2024, Jaslo Ziska
#
# Authors:
# Laurent Pinchart <laurent.pinchart@ideasonboard.com>
# Jaslo Ziska <jaslo@ziska.de>
#
# Generate gstreamer control properties from YAML

import argparse
import jinja2
import re
import sys
import yaml

from controls import Control


exposed_controls = [
    'AeEnable', 'AeMeteringMode', 'AeConstraintMode', 'AeExposureMode',
    'ExposureValue', 'ExposureTime', 'AnalogueGain', 'AeFlickerPeriod',
    'Brightness', 'Contrast', 'AwbEnable', 'AwbMode', 'ColourGains',
    'Saturation', 'Sharpness', 'ColourCorrectionMatrix', 'ScalerCrop',
    'DigitalGain', 'AfMode', 'AfRange', 'AfSpeed', 'AfMetering', 'AfWindows',
    'LensPosition', 'Gamma',
]


def find_common_prefix(strings):
    prefix = strings[0]

    for string in strings[1:]:
        while string[:len(prefix)] != prefix and prefix:
            prefix = prefix[:len(prefix) - 1]
        if not prefix:
            break

    return prefix


def format_description(description):
    # Substitute doxygen keywords \sa (see also) and \todo
    description = re.sub(r'\\sa((?: \w+)+)',
                         lambda match: 'See also: ' + ', '.join(
                             map(kebab_case, match.group(1).strip().split(' '))
                         ) + '.', description)
    description = re.sub(r'\\todo', 'Todo:', description)

    description = description.strip().split('\n')
    return '\n'.join([
        '"' + line.replace('\\', r'\\').replace('"', r'\"') + ' "' for line in description if line
    ]).rstrip()


# Custom filter to allow indenting by a string prior to Jinja version 3.0
#
# This function can be removed and the calls to indent_str() replaced by the
# built-in indent() filter when dropping Jinja versions older than 3.0
def indent_str(s, indention):
    s += '\n'

    lines = s.splitlines()
    rv = lines.pop(0)

    if lines:
        rv += '\n' + '\n'.join(
            indention + line if line else line for line in lines
        )

    return rv


def snake_case(s):
    return ''.join([
        c.isupper() and ('_' + c.lower()) or c for c in s
    ]).strip('_')


def kebab_case(s):
    return snake_case(s).replace('_', '-')


def extend_control(ctrl):
    if ctrl.vendor != 'libcamera':
        ctrl.namespace = f'{ctrl.vendor}::'
        ctrl.vendor_prefix = f'{ctrl.vendor}-'
    else:
        ctrl.namespace = ''
        ctrl.vendor_prefix = ''

    ctrl.is_array = ctrl.size is not None

    if ctrl.is_enum:
        # Remove common prefix from enum variant names
        prefix = find_common_prefix([enum.name for enum in ctrl.enum_values])
        for enum in ctrl.enum_values:
            enum.gst_name = kebab_case(enum.name.removeprefix(prefix))

        ctrl.gtype = 'enum'
        ctrl.default = '0'
    elif ctrl.element_type == 'bool':
        ctrl.gtype = 'boolean'
        ctrl.default = 'false'
    elif ctrl.element_type == 'float':
        ctrl.gtype = 'float'
        ctrl.default = '0'
        ctrl.min = '-G_MAXFLOAT'
        ctrl.max = 'G_MAXFLOAT'
    elif ctrl.element_type == 'int32_t':
        ctrl.gtype = 'int'
        ctrl.default = '0'
        ctrl.min = 'G_MININT'
        ctrl.max = 'G_MAXINT'
    elif ctrl.element_type == 'int64_t':
        ctrl.gtype = 'int64'
        ctrl.default = '0'
        ctrl.min = 'G_MININT64'
        ctrl.max = 'G_MAXINT64'
    elif ctrl.element_type == 'uint8_t':
        ctrl.gtype = 'uchar'
        ctrl.default = '0'
        ctrl.min = '0'
        ctrl.max = 'G_MAXUINT8'
    elif ctrl.element_type == 'Rectangle':
        ctrl.is_rectangle = True
        ctrl.default = '0'
        ctrl.min = '0'
        ctrl.max = 'G_MAXINT'
    else:
        raise RuntimeError(f'The type `{ctrl.element_type}` is unknown')

    return ctrl


def main(argv):
    # Parse command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--output', '-o', metavar='file', type=str,
                        help='Output file name. Defaults to standard output if not specified.')
    parser.add_argument('--template', '-t', dest='template', type=str, required=True,
                        help='Template file name.')
    parser.add_argument('input', type=str, nargs='+',
                        help='Input file name.')

    args = parser.parse_args(argv[1:])

    controls = {}
    for input in args.input:
        data = yaml.safe_load(open(input, 'rb').read())

        vendor = data['vendor']
        ctrls = controls.setdefault(vendor, [])

        for ctrl in data['controls']:
            ctrl = Control(*ctrl.popitem(), vendor, mode='controls')

            if ctrl.name in exposed_controls:
                ctrls.append(extend_control(ctrl))

    data = {'controls': list(controls.items())}

    env = jinja2.Environment()
    env.filters['format_description'] = format_description
    env.filters['indent_str'] = indent_str
    env.filters['snake_case'] = snake_case
    env.filters['kebab_case'] = kebab_case
    template = env.from_string(open(args.template, 'r', encoding='utf-8').read())
    string = template.render(data)

    if args.output:
        with open(args.output, 'w', encoding='utf-8') as output:
            output.write(string)
    else:
        sys.stdout.write(string)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# Copyright 2023 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from mojom_parser_test_case import MojomParserTestCase


class FeatureTest(MojomParserTestCase):
  """Tests feature parsing behavior."""
  def testFeatureOff(self):
    """Verifies basic parsing of feature types."""
    types = self.ExtractTypes("""
      // e.g. BASE_DECLARE_FEATURE(kFeature);
      [AttributeOne=ValueOne]
      feature kFeature {
        // BASE_FEATURE(kFeature,"MyFeature",
        //     base::FEATURE_DISABLED_BY_DEFAULT);
        const string name = "MyFeature";
        const bool default_state = false;
      };
    """)
    self.assertEqual('name', types['kFeature'].constants[0].mojom_name)
    self.assertEqual('"MyFeature"', types['kFeature'].constants[0].value)
    self.assertEqual('default_state', types['kFeature'].constants[1].mojom_name)
    self.assertEqual('false', types['kFeature'].constants[1].value)

  def testFeatureOn(self):
    """Verifies basic parsing of feature types."""
    types = self.ExtractTypes("""
      // e.g. BASE_DECLARE_FEATURE(kFeature);
      feature kFeature {
        // BASE_FEATURE(kFeature,"MyFeature",
        //     base::FEATURE_ENABLED_BY_DEFAULT);
        const string name = "MyFeature";
        const bool default_state = true;
      };
    """)
    self.assertEqual('name', types['kFeature'].constants[0].mojom_name)
    self.assertEqual('"MyFeature"', types['kFeature'].constants[0].value)
    self.assertEqual('default_state', types['kFeature'].constants[1].mojom_name)
    self.assertEqual('true', types['kFeature'].constants[1].value)

  def testFeatureWeakKeyword(self):
    """Verifies that `feature` is a weak keyword."""
    types = self.ExtractTypes("""
      // e.g. BASE_DECLARE_FEATURE(kFeature);
      [AttributeOne=ValueOne]
      feature kFeature {
        // BASE_FEATURE(kFeature,"MyFeature",
        //     base::FEATURE_DISABLED_BY_DEFAULT);
        const string name = "MyFeature";
        const bool default_state = false;
      };
      struct MyStruct {
         bool feature = true;
      };
      interface InterfaceName {
         Method(string feature) => (int32 feature);
      };
    """)
    self.assertEqual('name', types['kFeature'].constants[0].mojom_name)
    self.assertEqual('"MyFeature"', types['kFeature'].constants[0].value)
    self.assertEqual('default_state', types['kFeature'].constants[1].mojom_name)
    self.assertEqual('false', types['kFeature'].constants[1].value)

  def testFeatureAttributesAreFeatures(self):
    """Verifies that feature values in attributes are really feature types."""
    a_mojom = 'a.mojom'
    self.WriteFile(
        a_mojom, 'module a;'
        'feature F { const string name = "f";'
        'const bool default_state = false; };')
    b_mojom = 'b.mojom'
    self.WriteFile(
        b_mojom, 'module b;'
        'import "a.mojom";'
        'feature G'
        '{const string name = "g"; const bool default_state = false;};'
        '[Attri=a.F] interface Foo { Foo(); };'
        '[Boink=G] interface Bar {};')
    self.ParseMojoms([a_mojom, b_mojom])
    b = self.LoadModule(b_mojom)
    self.assertEqual(b.interfaces[0].attributes['Attri'].mojom_name, 'F')
    self.assertEqual(b.interfaces[1].attributes['Boink'].mojom_name, 'G')
